""" auxiliary file """
from qiskit import execute

def measurement(circuit, n_qubits, classical_register, backend, shots):
    """ run circuit and return measurements """
    circuit.measure(list(range(n_qubits)), classical_register)

    job = execute(circuit, backend, shots=shots, optimization_level=3)

    counts = job.result().get_counts(circuit)

    counts2 = {}
    for k in range(2 ** n_qubits):
        pattern = '{:0{}b}'.format(k, n_qubits)
        if pattern in counts:
            counts2[pattern] = counts[pattern]
        else:
            counts2[pattern] = 0.0

    return [value / shots for (key, value) in counts2.items()]



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Maximally multipartite entangled state (MMES).
https://journals.aps.org/pra/abstract/10.1103/PhysRevA.77.060304
https://link.springer.com/article/10.1007/s10773-018-3956-3
"""

from itertools import combinations
import numpy as np

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

def create_state(n_qubits):
    bell_phi_plus  = np.array([1, 0, 0, 1]).reshape((4, 1)) # 1/sqrt(2) (|00>+|11>)
    bell_phi_minus = np.array([1, 0, 0, -1]).reshape((4, 1)) # 1/sqrt(2) (|00>-|11>)
    #bell_psi_plus  = np.array([0, 1, 1, 0]).reshape((4, 1)) # 1/sqrt(2) (|01>+|10>)
    #bell_psi_minus  = np.array([0, -1, 1, 0]).reshape((4, 1)) # 1/sqrt(2) (-|01>+|10>)

    #psi = [bell_phi_plus, bell_psi_plus, bell_phi_minus, bell_psi_minus]
    #psi = [bell_phi_plus, bell_psi_minus, bell_psi_plus, bell_phi_minus]
    #psi = [bell_phi_plus, bell_psi_minus] # , bell_psi_plus, bell_phi_minus]
    psi = [bell_phi_plus, bell_phi_minus] # , bell_psi_plus, bell_phi_minus]

    return _create_state(n_qubits, psi) #, bell_psi_plus)
    #return _create_state(n_qubits, bell_phi_plus, bell_phi_minus, bell_psi_plus)

def _create_state(n_qubits, psi): #, psi2): #, psi3):
    #if not _ortho(psi):
    #    raise ValueError(
    #        "The state vectors do not satisfy the orthogonality condition."
    #    )

    previous_n = int(np.log2(psi[0].shape[0]))
    if previous_n == n_qubits:
        vector = psi[0].reshape(2**n_qubits)
        return vector/np.linalg.norm(vector)

    eye  = np.eye(2)
    zero = eye[:,[0]]
    one  = eye[:,[1]]

    #for qubit in range(previous_n):
    #    psi4 = _pauli_x(qubit, previous_n).dot(psi3)
    #    if _ortho(psi1, psi4) and _ortho(psi2, psi4) and _ortho(psi3, psi4):
    #        print('break')
    #        break
    #    if qubit > 0:
    #        psi4 = _pauli_z(qubit-1, previous_n).dot(psi4)
    #    if _ortho(psi1, psi4) and _ortho(psi2, psi4) and _ortho(psi3, psi4):
    #        print('break')
    #        break

    #signal = (-1)**previous_n
    new_psi = []
    for pair in combinations(psi, 2):
        #for signal in [(-1)**previous_n, (-1)**(previous_n+1)]:
        vector = np.kron(pair[0], zero) + np.kron(pair[1], one)
        if _ortho([*new_psi, vector]):
            new_psi.append(vector)
        vector = np.kron(pair[0], zero) - np.kron(pair[1], one)
        if _ortho([*new_psi, vector]):
            new_psi.append(vector)
        """
        vector = np.kron(pair[1], zero) + np.kron(pair[0], one)
        if _ortho([*new_psi, vector]):
            new_psi.append(vector)
        vector = np.kron(pair[1], zero) - np.kron(pair[0], one)
        if _ortho([*new_psi, vector]):
            new_psi.append(vector)
        """

    psi.clear()
    return _create_state(n_qubits, new_psi) # , new_psi2)

def _count(vector):
    return sum([abs(i) > 10**-15 for i in vector])

def _ortho(vectors):
    if len(vectors) == 1:
        return True
    for pair in combinations(vectors, 2):
        if np.abs(_braket(pair[0], pair[1])) > 10**-15:
            return False
    return True

def _braket(vector1, vector2):
    return np.conj(vector1.T).dot(vector2)

def _pauli_z(qubit, n_qubits):
    return _pauli(qubit, n_qubits, [[1,0], [0,-1]])

def _pauli_x(qubit, n_qubits):
    return _pauli(qubit, n_qubits, [[0,1], [1,0]])

def _pauli(qubit, n_qubits, pauli):
    i_gate = np.identity(2)
    gates = [[1]]
    for i in range(n_qubits):
        if i == qubit:
            gates = np.kron(gates, pauli)
        else:
            gates = np.kron(gates, i_gate)

    return gates



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the mottonen.py module.
"""

from unittest import TestCase
import numpy as np
from qiskit import ClassicalRegister, execute
from qiskit.providers.aer.backends import AerSimulator
from qclib.state_preparation.mottonen import initialize
from qclib.util import get_state

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

backend = AerSimulator()
SHOTS = 8192

class TestMottonen(TestCase):
    @staticmethod
    def measurement(circuit, n_qubits, classical_reg):
        circuit.measure(list(range(n_qubits)), classical_reg)

        job = execute(circuit, backend, shots=SHOTS, optimization_level=3)

        counts = job.result().get_counts(circuit)
        sum_values = sum(counts.values())

        counts2 = {}
        for i in range(2**n_qubits):
            pattern = '{:0{}b}'.format(i, n_qubits)
            if pattern in counts:
                counts2[pattern] = counts[pattern]
            else:
                counts2[pattern] = 0.0

        return [ value/sum_values for (key, value) in counts2.items() ]

    @staticmethod
    def mottonen_experiment(state):
        circuit = initialize(state)

        n_qubits = int(np.log2(len(state)))
        classical_reg = ClassicalRegister(n_qubits)
        circuit.add_register(classical_reg)

        return TestMottonen.measurement(circuit, n_qubits, classical_reg)

    def test_mottonen_state_real(self):
        state_vector = np.random.rand(32)
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state))

    def test_mottonen_state_complex(self):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state))

    def test_mottonen_measure(self):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        state = TestMottonen.mottonen_experiment(state_vector)

        self.assertTrue(np.allclose( np.power(np.abs(state_vector),2), state,
                        rtol=1e-01, atol=0.005))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test sparse initialization """

from unittest import TestCase
import numpy as np
from qclib.state_preparation.cvoqram import cvoqram_initialize
from qclib.util import double_sparse, get_state
from qclib.state_preparation.pivot import PivotStatePreparation
pivot_initialize = PivotStatePreparation.initialize


class TestCvoqram(TestCase):
    """ Testing qclib.state_preparation.cvoqram """
    def test_cvoqram(self):
        """ Testing cvoqram 2 amplitudes and with auxiliary qubits """
        data = [([0, 0, 1], 1/np.sqrt(3)), ([1, 1, 0], np.sqrt(2j/3))]
        qc_cvoqram = cvoqram_initialize(data)
        state = get_state(qc_cvoqram)
        self.assertTrue(np.isclose(state[0b110000], np.sqrt(2j/3)))
        self.assertTrue(np.isclose(state[0b001000], np.sqrt(1/3)))

    def test_cvoqram_without_aux(self):
        """ Testing cvoqram 2 amplitudes and with auxiliary qubits """
        data = [([0, 0, 1], 1j * np.sqrt(1/3)), ([1, 1, 0], np.sqrt(2j/3))]
        qc_cvoqram = cvoqram_initialize(data, False)
        state = get_state(qc_cvoqram)
        self.assertTrue(np.isclose(state[0b1100], np.sqrt(2j/3)))
        self.assertTrue(np.isclose(state[0b0010], 1j * np.sqrt(1/3)))

    def test_cvoqram_random(self):
        """ Testing cvoqram 4 amplitudes and with auxiliary qubits """
        n_qubits = 4
        log_npatterns = 2
        prob = 0.2
        data = double_sparse(n_qubits, log_npatterns, prob)
        qc_cvoqram = cvoqram_initialize(data)
        state = get_state(qc_cvoqram)
        for k, _ in enumerate(data):
            lst = data[k][0]
            bin_index = ''.join(map(str, lst))
            bin_index = bin_index + n_qubits * '0'  # padding work qubits

            self.assertTrue(np.isclose(state[int(bin_index, 2)], data[k][1]))

    def test_double_sparse(self):
        """ Test double sparse random generation """
        n_qubits = 6
        log_npatterns = 4
        prob = 0.8
        data = double_sparse(n_qubits, log_npatterns, prob)
        norm = 0
        for k, _ in enumerate(data):
            norm = norm + np.abs(data[k][1])**2
            self.assertTrue(np.sum(data[k][0]) <= np.sum(data[k][0]))

        self.assertTrue(np.isclose(norm, 1))


class TestPivotingSP(TestCase):
    """ Testing pivot state preparation """
    def test_pivoting_sp(self):
        """ Testing pivot state preparation with 2 amplitudes """
        data = {'001': 1 / np.sqrt(3), '110': np.sqrt(2j / 3)}
        circuit = pivot_initialize(data)
        state = get_state(circuit)
        self.assertTrue(np.isclose(state[0b001], 1/np.sqrt(3)))
        self.assertTrue(np.isclose(state[0b110], np.sqrt(2j / 3)))

    def test_pivotsp_random(self):
        """ Testing pivot state preparation with 4 amplitudes """
        n_qubits = 4
        log_npatterns = 2
        prob = 0.2
        data = double_sparse(n_qubits, log_npatterns, prob)
        dict_data = {}
        for value in data:
            index = ''.join(map(str, value[0]))
            dict_data[index] = value[1]

        circuit = pivot_initialize(dict_data)
        state = get_state(circuit)
        for k, _ in enumerate(data):
            lst = data[k][0]
            bin_index = ''.join(map(str, lst))

            self.assertTrue(np.isclose(state[int(bin_index, 2)], data[k][1]))

    def test_pivotsp_random_aux(self):
        """ Testing pivot state preparation with 4 amplitudes and auxiliary qubits"""
        n_qubits = 4
        log_npatterns = 2
        prob = 0.2
        data = double_sparse(n_qubits, log_npatterns, prob)
        dict_data = {}
        for value in data:
            index = ''.join(map(str, value[0]))
            dict_data[index] = value[1]

        circuit = pivot_initialize(dict_data, aux=True)
        state = get_state(circuit)
        for k, _ in enumerate(data):
            lst = data[k][0]
            bin_index = ''.join(map(str, lst))
            bin_index = bin_index + '0'

            self.assertTrue(np.isclose(state[int(bin_index, 2)], data[k][1]))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Gleining et al algorithm for creating a quantum
circuit that loads a sparse state.
https://ieeexplore.ieee.org/document/9586240
'''

import numpy as np
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit.library import U3Gate


def initialize(state):
    """
    Classical algorithm that creates a quantum circuit C that loads
    a sparse quantum state, applying a sequence of operations maping
    the desired state |sigma> to |0>. And then inverting C to obtain
    the mapping of |0> to the desired state |sigma>.
    Args:
      state: Sparse array with real values
    Returns:
      Quantum circuit C^{-1} that maps |0> to the desired state |sigma>
    """
    state_dict = _build_state_dict(state)

    b_strings = list(state_dict.keys())

    n_qubits = len(b_strings[0])
    quantum_register = QuantumRegister(n_qubits)
    quantum_circuit = QuantumCircuit(quantum_register)

    while len(b_strings) > 1:
        state_dict, quantum_circuit = _merging_procedure(
            state_dict, quantum_circuit)
        b_strings = list(state_dict.keys())

    b_string = b_strings.pop()
    for (bit_idx, bit) in enumerate(b_string):
        if bit == '1':
            quantum_circuit.x(bit_idx)

    return quantum_circuit.reverse_ops()


def _build_state_dict(state):
    """
    Builds a dict of the non zero amplitudes with their
    associated binary strings as follows:
      { '000': <value>, ... , '111': <value> }
    Args:
      state: The classical description of the state vector
    """
    n_qubits = np.ceil(np.log(len(state))).astype(int)
    state_dict = {}
    for (value_idx, value) in enumerate(state):
        if value != 0:
            binary_string = '{:0{}b}'.format(value_idx, n_qubits)[::-1]
            state_dict[binary_string] = value
    return state_dict


def _maximizing_difference_bit_search(b_strings, dif_qubits):
    """
    Splits the set of bit strings into two (t_0 and t_1), by setting
    t_0 as the set of bit_strings with 0 in the b-th position, and
    t_1 as the set of bit_strings with 1 in the b-th position.
    Searching for the bit b that maximizes the difference between
    t_0 and t_1 where neither is empty.
    Args:
      b_string: A list of bit strings eg.: ['000', '011', ...,'101']
      dif_qubits: A list of previous qubits found to maximize the difference
    Returns:
      bit_index: The qubit index that maximizes the splitting of the list b_strings
      t_0: List of binary strings with 0 on the b-th qubit
      t_1: List of binary strings with 1 on the b-th qubit
    """
    t_0 = []
    t_1 = []
    bit_index = 0
    set_difference = 0
    bit_search_space = list(set(range(len(b_strings[0]))) - set(dif_qubits))
    for bit in bit_search_space:
        temp_t0 = []
        temp_t1 = []
        for bit_string in b_strings:
            if bit_string[bit] == '0':
                temp_t0.append(bit_string)
            else:
                temp_t1.append(bit_string)
        # Neither temp_t0 nor temp_t1 must be empty
        if (temp_t0 and temp_t1):
            temp_difference = np.abs(len(temp_t0) - len(temp_t1))
            if temp_difference == 0 and not t_0 and not t_1:
                t_0 = temp_t0
                t_1 = temp_t1
                bit_index = bit
            elif temp_difference > set_difference:
                t_0 = temp_t0
                t_1 = temp_t1
                bit_index = bit
                set_difference = temp_difference
    return bit_index, t_0, t_1


def _build_bit_string_set(b_strings, dif_qubits, dif_values):
    """
    Creates a new set of bit strings from b_strings, where the bits
    in the indexes in dif_qubits match the values in dif_values.

    Args:
      b_strings: list of bit strings eg.: ['000', '011', ...,'101']
      dif_qubits: list of integers with the bit indexes
      dif_values: list of integers values containing the values each bit
                  with index in dif_qubits shoud have
    Returns:
      A new list of bit_strings, with matching values in dif_values
      on indexes dif_qubits
    """
    bit_string_set = []

    for b_string in b_strings:
        include_string = True
        for (b_index, b_value) in zip(dif_qubits, dif_values):
            if b_string[b_index] != b_value:
                include_string = False
        if include_string:
            bit_string_set.append(b_string)
    return bit_string_set


def _bit_string_search(b_strings, dif_qubits, dif_values):
    """
    Searches for the bit strings with unique qubit values in `dif_values`
    on indexes `dif_qubits`.
    Args:
      b_strings: List of binary strings where the search is to be performed
                  e.g.: ['000', '010', '101', '111']
      dif_qubits: List of indices on a binary string of size N e.g.: [1, 3, 5]
      dif_values: List of values each qubit must have on indexes stored in dif_qubits [0, 1, 1]
    Returns:
      b_strings: One size list with the string found, to have values dif_values on indexes
                  dif_qubits
      dif_qubits: Updated list with new indexes
      dif_values: Updated list with new values
    """
    temp_strings = b_strings
    while len(temp_strings) > 1:
        bit, t_0, t_1 = _maximizing_difference_bit_search(
            temp_strings, dif_qubits)
        dif_qubits.append(bit)
        if len(t_0) < len(t_1):
            dif_values.append('0')
            temp_strings = t_0
        else:
            dif_values.append('1')
            temp_strings = t_1
        # dif_qubits must have at least two values stored in it
        if len(temp_strings) == 1 and len(dif_qubits) == 1:
            temp_strings = b_strings

    return temp_strings, dif_qubits, dif_values


def _search_bit_strings_for_merging(state_dict):
    """
    Searches for the states described by the bit strings bitstr1 and bitstr2 to be merged
    Args:
      state_dict: A dictionary with the non-zero amplitudes associated to their corresponding
                  binary strings as keys e.g.: {'001': <value>, '101': <value>}
    Returns:
      bitstr1: First binary string
      bitstr2: Second binary string
      dif_qubit: Qubit index to be used as target for the merging operation
      dif_qubits: List of qubit indexes where bitstr1 and bitstr2 must be equal, because the
                  correspondig qubits of those indexes are to be used as control for the
                  merging operation
    """
    # Initialization
    dif_qubits = []
    dif_values = []
    b_strings1 = b_strings2 = list(state_dict.keys())
    bitstr1 = None
    bitstr2 = None

    if len(b_strings1) == 2:
        # Search for the difference bit
        bit, t_0, t_1 = _maximizing_difference_bit_search(b_strings1, dif_qubits)
        dif_qubit = bit
        bitstr1 = t_1[0]
        bitstr2 = t_0[0]
    else:
        # Searching for bitstr1
        b_strings1, dif_qubits, dif_values = _bit_string_search(b_strings1, dif_qubits, dif_values)
        dif_qubit = dif_qubits.pop()
        dif_values.pop()
        bitstr1 = b_strings1[0]

        # Searching for bitstr2
        b_strings1 = _build_bit_string_set(b_strings2, dif_qubits, dif_values)
        b_strings1, dif_qubits, dif_values = _bit_string_search(b_strings1, dif_qubits, dif_values)
        bitstr2 = b_strings1[0]

    return bitstr1, bitstr2, dif_qubit, dif_qubits


def _apply_operation_to_bit_string(b_string, operation, qubit_indexes):
    """
    Applies changes on binary strings according to the operation
    Args:
      b_string: Binary string '00110'
      operation: Operation to be applied to the string
      qubit_indexes: Indexes of the qubits on the binary strings where the operations are to
                      be applied
    Returns:
      Updated binary string
    """
    assert operation in ['x', 'cx']
    compute_op = None
    if operation == 'x':
        compute_op = lambda x, idx : \
                     x[:idx] + '1' + x[idx+1:] \
                     if x[idx] == '0' else x[0:idx] + '0' + x[idx+1:]
    elif operation == 'cx':
        compute_op = lambda x, idx: \
                     x[:idx[1]] + "{}".format((not int(x[idx[1]])) * 1) + x[idx[1]+1:] \
                     if x[idx[0]] == '1' else x

    return compute_op(b_string, qubit_indexes)

def _update_state_dict_according_to_operation(state_dict, operation, qubit_indexes,
                                                                     merge_strings=None):
    """
    Updates the keys of the state_dict according to the operation being applied to the circuit
    Args:
      state_dict: A dictionary with the non-zero amplitudes associated to their corresponding
                  binary strings as keys e.g.: {'001': <value>, '101': <value>}
      operation: Operation to be applied to the states, it must be ['x', 'cx', 'merge']
      qubit_indexes: Indexes of the qubits on the binary strings where the operations are to
                      be applied
      merge_strings: Binary strings associated ot the states on the quantum processor
                      to be merge e.g.:['01001', '10110']
    Returns:
      A state_dict with the updated states
    """
    assert operation in ['x', 'cx', 'merge']
    state_list = list(state_dict.items())
    new_state_dict = {}
    if operation == 'merge':
        assert merge_strings is not None
        # Computes the norm of bitstr1 and bitstr2
        new_state_dict = state_dict.copy()
        norm = np.linalg.norm(
            [new_state_dict[merge_strings[0]], new_state_dict[merge_strings[1]]])
        new_state_dict.pop(merge_strings[1], None)
        new_state_dict[merge_strings[0]] = norm
    else:
        for (bit_string, value) in state_list:
            temp_bstring = _apply_operation_to_bit_string(
                bit_string, operation, qubit_indexes)
            new_state_dict[temp_bstring] = value

    return new_state_dict


def _equalize_bit_string_states(bitstr1, bitstr2, dif, state_dict, quantum_circuit):
    """
    Applies operations to the states represented by bit strings bitstr1 and bitstr2 equalizing them
    at every qubit except the one in the dif index. And alters the bit strings and state_dict
    accordingly.
    Args:
      bitstr1: First bit string
      bitstr2: Second bit string
      dif: index where both bitstr1 and bitstr2 mus be different
      state_dict: A dictionary with the non-zero amplitudes associated to their corresponding
                  binary strings as keys e.g.: {'001': <value>, '101': <value>}
      quantum_circuit: Qiskit's quantum circuit's object with the gates applied to the circuit
    Returns:
      Updated bitstr1, bitstr2, state_dict and quantum_circuit
    """
    b_index_list = list(range(len(bitstr1)))
    b_index_list.remove(dif)

    for b_index in b_index_list:
        if bitstr1[b_index] != bitstr2[b_index]:
            quantum_circuit.cx(dif, b_index)
            bitstr1 = _apply_operation_to_bit_string(bitstr1, 'cx', [dif, b_index])
            bitstr2 = _apply_operation_to_bit_string(bitstr2, 'cx', [dif, b_index])
            state_dict = _update_state_dict_according_to_operation(
                state_dict, 'cx', [dif, b_index])

    return bitstr1, bitstr2, state_dict, quantum_circuit


def _apply_not_gates_to_qubit_index_list(bitstr1, bitstr2, dif_qubits, state_dict, quantum_circuit):
    """
    Applies quantum not gate at the qubit at a given index, where the state represented by the
    bit string bitstr2 is different than '1' at index in diff_qubits.
    Args:
      bitstr1: First bit string
      bitstr2: Second bit string
      dif_qubits: indexes where both bitstr1 and bitstr2 are equal
      state_dict: A dictionary with the non-zero amplitudes associated to their corresponding
                  binary strings as keys e.g.: {'001': <value>, '101': <value>}
      quantum_circuit: Qiskit's quantum circuit's object with the gates applied to the circuit
    Returns:
      Updated bitstr1, bitstr2, state_dict and quantum_circuit
    """
    for b_index in dif_qubits:
        if bitstr2[b_index] != '1':
            quantum_circuit.x(b_index)
            bitstr1 = _apply_operation_to_bit_string(bitstr1, 'x', b_index)
            bitstr2 = _apply_operation_to_bit_string(bitstr2, 'x', b_index)
            state_dict = _update_state_dict_according_to_operation(
                state_dict, 'x', b_index)
    return bitstr1, bitstr2, state_dict, quantum_circuit


def _preprocess_states_for_merging(bitstr1, bitstr2, dif, dif_qubits, state_dict, quantum_circuit):
    """
    Apply the operations on the basis states to prepare for merging bitstr1 and bitstr2.
    Args:
      state_dict: A dictionary with the non-zero amplitudes associated to their corresponding
                  binary strings as keys e.g.: {'001': <value>, '101': <value>}
      bitstr1: First binary string to be merged
      bitstr2: Second binary string to be merged
      dif_qubits: List of qubit indexes on the binary strings
      dif: Target qubit index where the merge operation is to be applied
      quantum_circuit: Qiskit's QuantumCircuit object where the operations are to be called
    Returns:
      state_dict: Updated state dict
      bitstr1: First updated binary string to be merge
      bitstr2: Second updated binary string to be merged
      quantum_circuit: Qiskit's quantum circuit's object with the gates applied to the circuit
    """

    if bitstr1[dif] != '1':
        quantum_circuit.x(dif)
        bitstr1 = _apply_operation_to_bit_string(bitstr1, 'x', dif)
        bitstr2 = _apply_operation_to_bit_string(bitstr2, 'x', dif)
        state_dict = _update_state_dict_according_to_operation(
            state_dict, 'x', dif)

    bitstr1, bitstr2, state_dict, quantum_circuit = _equalize_bit_string_states(
        bitstr1, bitstr2, dif, state_dict, quantum_circuit)

    bitstr1, bitstr2, state_dict, quantum_circuit = _apply_not_gates_to_qubit_index_list(
        bitstr1, bitstr2, dif_qubits, state_dict, quantum_circuit)

    return bitstr1, bitstr2, state_dict, quantum_circuit


def _compute_angles(amplitude_1, amplitude_2):
    """
    Computes the angles for the adjoint of the merge matrix M
    that is going to map the dif qubit to zero e.g.:
      M(a|0> + b|1>) -> |0>

    Args:
      amplitude_1: A complex/real value, associated with the string with
                    1 on the dif qubit
      amplitude_2: A complex/real value, associated with the string with
                    0 on the dif qubit
    Returns:
      The angles theta, lambda and phi for the U3 operator
    """
    norm = np.linalg.norm([amplitude_1, amplitude_2])

    phi = 0
    lamb = 0
    # there is no minus on the theta because the intetion is to compute the inverse
    if isinstance(amplitude_1, complex) or isinstance(amplitude_2, complex):
        amplitude_1 = complex(amplitude_1) if not isinstance(
            amplitude_1, complex) else amplitude_1
        amplitude_2 = complex(amplitude_2) if not isinstance(
            amplitude_2, complex) else amplitude_2

        theta = - 2 * np.arcsin(np.abs(amplitude_2 / norm))
        lamb = np.log(amplitude_2 / norm).imag
        phi = np.log(amplitude_1 / norm).imag - lamb

    else:
        theta = - 2 * np.arcsin(amplitude_2 / norm)

    return theta, phi, lamb


def _merging_procedure(state_dict, quantum_circuit):

    bitstr1, bitstr2, dif, dif_qubits = _search_bit_strings_for_merging(state_dict)

    # Circuit building
    bitstr1, bitstr2, state_dict, quantum_circuit = _preprocess_states_for_merging(
        bitstr1, bitstr2, dif, dif_qubits, state_dict, quantum_circuit
    )

    theta, phi, lamb = _compute_angles(state_dict[bitstr1], state_dict[bitstr2])

    # Applying merge operation
    merge_gate = None
    if not dif_qubits:
        merge_gate = U3Gate(theta, phi, lamb, label='U3')
    else:
        merge_gate = U3Gate(theta, phi, lamb, label='U3').control(
            num_ctrl_qubits=len(dif_qubits))

    quantum_circuit.append(merge_gate, dif_qubits+[dif], [])
    state_dict = _update_state_dict_according_to_operation(
        state_dict, 'merge', None, merge_strings=[bitstr1, bitstr2]
    )

    return state_dict, quantum_circuit



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This module provides access to functions to
implement generic quantum computations.
'''

# pylint: disable=maybe-no-member

from math import ceil, log2
import scipy as sp
import qiskit
from qiskit import transpile
import numpy as np

def unitary(gate, decomposition='csd'):
    """
    Implements a generic quantum computation from a
    unitary matrix gate using the cosine sine decomposition.
    """
    size = len(gate)
    if size > 4:
        n_qubits = int(log2(size))

        qubits = qiskit.QuantumRegister(n_qubits)
        circuit = qiskit.QuantumCircuit(qubits)

        right_gates, theta, left_gates = \
            sp.linalg.cossin(gate, size/2, size/2, separate=True)

        gate_left = _unitary(list(left_gates), n_qubits, decomposition)
        gate_right = _unitary(list(right_gates), n_qubits, decomposition)

        circuit = circuit.compose(gate_left, qubits)
        circuit.ucry(list(2*theta), list(range(n_qubits-1)), n_qubits-1)
        circuit = circuit.compose(gate_right, qubits)

        return circuit

    circuit = qiskit.QuantumCircuit(log2(size))
    circuit.unitary(gate, circuit.qubits)

    return circuit

def _unitary(gate_list, n_qubits, decomposition='csd'):

    if len(gate_list[0]) == 2:
        qubits = qiskit.QuantumRegister(n_qubits)
        circuit = qiskit.QuantumCircuit(qubits)
        circuit.uc(gate_list, qubits[1:], qubits[0])
        return circuit

    if decomposition == 'csd':
        return _csd(gate_list, n_qubits)

    # QSD
    return _qsd(*gate_list)



# CSD decomposition



def _csd(gate_list, n_qubits):
    left, mid, right = _multiplexed_csd(gate_list)

    gate_left = _unitary(left, n_qubits)
    gate_right = _unitary(right, n_qubits)

    qubits = qiskit.QuantumRegister(n_qubits)
    circuit = qiskit.QuantumCircuit(qubits)

    circuit = circuit.compose(gate_left, qubits)

    target = int(n_qubits - log2(len(left)))
    control = list(range(0, target)) + list(range(target + 1, n_qubits))
    circuit.ucry(list(mid), control, target)

    circuit = circuit.compose(gate_right, qubits)

    return circuit

def _multiplexed_csd(gate_list):
    left = []
    mid = []
    right = []
    size = len(gate_list[0])
    for gate in gate_list:
        right_gates, theta, left_gates = \
            sp.linalg.cossin(gate, size / 2, size / 2, separate=True)

        left = left + list(left_gates)
        right = right + list(right_gates)
        mid = mid + list(2 * theta)

    return left, mid, right



# QSD decomposition



def _qsd(gate1, gate2):
    n_qubits = int(log2(len(gate1))) + 1

    list_d, gate_v, gate_w = _compute_gates(gate1, gate2)

    left_gate = unitary(gate_w, 'qsd')
    right_gate = unitary(gate_v, 'qsd')

    qubits = qiskit.QuantumRegister(n_qubits)
    circuit = qiskit.QuantumCircuit(qubits)

    circuit = circuit.compose(left_gate, qubits[0:-1])
    circuit.ucrz(list(-2*np.angle(list_d)), qubits[0:-1], qubits[-1])
    circuit = circuit.compose(right_gate, qubits[0:-1])
    return circuit

def _is_unitary(matrix):
    is_identity = np.conj(matrix.T).dot(matrix)
    return np.allclose(is_identity, np.identity(matrix.shape[0]))

def _closest_unitary(matrix):
    svd_u,_,svd_v = np.linalg.svd(matrix)
    return svd_u.dot(svd_v)

def _compute_gates(gate1, gate2):

    d_square, gate_v = np.linalg.eig(gate1 @ gate2.conj().T)
    list_d = np.sqrt(d_square, dtype=complex)
    gate_d = np.diag(list_d)

    if not _is_unitary(gate_v):
        # degeneracy
        gate_v = _closest_unitary(gate_v)

    gate_w = gate_d @ gate_v.conj().T @ gate2

    return list_d, gate_v, gate_w



# CNOT count



def cnot_count(gate, decomposition='csd', method='estimate'):
    """
    Count the number of CNOTs to decompose the unitary.
    """
    if method == 'estimate':
        return _cnot_count_estimate(gate, decomposition)

    # Exact count
    circuit = unitary(gate, decomposition)
    transpiled_circuit = transpile(circuit, basis_gates=['u1','u2','u3','cx'],
                                                            optimization_level=0)
    count_ops = transpiled_circuit.count_ops()
    if 'cx' in count_ops:
        return count_ops['cx']

    return 0

def _cnot_count_estimate(gate, decomposition='csd'):
    """
    Estimate the number of CNOTs to decompose the unitary.
    """
    n_qubits = int(log2(gate.shape[0]))
    if n_qubits == 1:
        return 0

    if decomposition == 'csd':
        # Table 1 from "Synthesis of Quantum Logic Circuits", Shende et al.
        return int( ceil( 4**n_qubits - 2*2**n_qubits ) )

    # Upper-bound expression for the unitary decomposition QSD l=2 without the optimizations.
    # With the optimizations, they need to be replaced.
    # Table 1 from "Synthesis of Quantum Logic Circuits", Shende et al.
    return int( ceil( 9/16*2**(2*n_qubits) - 3/2 * 2**n_qubits ) )



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implements the state preparation
defined at https://arxiv.org/abs/1003.5760.
"""

from math import ceil, log2
import numpy as np
import deprecation
from qiskit import QuantumCircuit, QuantumRegister
from qclib.state_preparation.mottonen import initialize as mottonen
from qclib.unitary import unitary as decompose_unitary, cnot_count as cnots_unitary
from qclib.isometry import decompose as decompose_isometry, cnot_count as cnots_isometry

# pylint: disable=maybe-no-member

def initialize(state_vector, low_rank=0, isometry_scheme='ccd', unitary_scheme='qsd',
                                                                        partition=None):
    """
    Low-rank state preparation using Schmidt decomposition.
    https://arxiv.org/abs/2111.03132

    For instance, to initialize the state a|00> + b|10> (|a|^2+|b|^2=1)
        $ state = [a, 0, b, 0]
        $ circuit = initialize(state)

    Parameters
    ----------
    state_vector: list of complex
        A unit vector representing a quantum state.
        Values are amplitudes.

    low_rank: int
        ``state`` low-rank approximation (1 <= ``low_rank`` < 2**(n_qubits//2)).
        If ``low_rank`` is not in the valid range, it will be ignored.
        This parameter limits the rank of the Schmidt decomposition. If the Schmidt rank
        of the state decomposition is greater than ``low_rank``, a low-rank approximation
        is applied.

    isometry_scheme: string
        Scheme used to decompose isometries.
        Possible values are ``'knill'`` and ``'ccd'`` (column-by-column decomposition).
        Default is ``isometry_scheme='ccd'``.

    unitary_scheme: string
        Scheme used to decompose unitaries.
        Possible values are ``'csd'`` (cosine-sine decomposition) and ``'qsd'`` (quantum
        Shannon decomposition).
        Default is ``unitary_scheme='qsd'``.

    partition: list of int
        Set of qubit indices that represent a part of the bipartition.
        The other partition will be the relative complement of the full set of qubits
        with respect to the set ``partition``.
        The valid range for indexes is ``0 <= index < n_qubits``. The number of indexes
        in the partition must be greater than or equal to ``1`` and less than or equal
        to ``n_qubits//2`` (``n_qubits//2+1`` if ``n_qubits`` is odd).
        Default is ``partition=list(range(n_qubits//2 + odd))``.

    Returns
    -------
    circuit: QuantumCircuit
        QuantumCircuit to initialize the state.
    """

    n_qubits = _to_qubits(len(state_vector))
    if n_qubits < 2:
        return mottonen(state_vector)

    circuit, reg_a, reg_b = _create_quantum_circuit(state_vector, partition)

    # Schmidt decomposition
    svd_u, singular_values, svd_v = schmidt_decomposition(state_vector, reg_a)

    rank, svd_u, svd_v, singular_values = \
        low_rank_approximation(low_rank, svd_u, svd_v, singular_values)

    # Schmidt measure of entanglement
    ebits = _to_qubits(rank)

    # Phase 1. Encodes the singular values.
    if ebits > 0:
        reg_sv = reg_b[:ebits]
        singular_values = singular_values / np.linalg.norm(singular_values)
        _encode(singular_values.reshape(rank, 1), circuit, reg_sv,
                                            isometry_scheme, unitary_scheme)

    # Phase 2. Entangles only the necessary qubits, according to rank.
    for j in range(ebits):
        circuit.cx(reg_b[j], reg_a[j])

    # Phase 3 and 4 encode gates U and V.T
    _encode(svd_u, circuit, reg_b, isometry_scheme, unitary_scheme)
    _encode(svd_v.T, circuit, reg_a, isometry_scheme, unitary_scheme)

    return circuit.reverse_bits()



# Schmidt routines



def schmidt_decomposition(state_vector, partition):
    """
    Execute the Schmidt decomposition of a state vector.

    Parameters
    ----------
    state_vector: list of complex
        A unit vector representing a quantum state.
        Values are amplitudes.

    partition: list of int
        Set of qubit indices that represent a part of the bipartition.
        The other partition will be the relative complement of the full set of qubits
        with respect to the set ``partition``.
        The valid range for indexes is ``0 <= index < n_qubits``. The number of indexes
        in the partition must be greater than or equal to ``1`` and less than or equal
        to ``n_qubits//2`` (``n_qubits//2+1`` if ``n_qubits`` is odd).
    """

    n_qubits = _to_qubits(len(state_vector))

    sep_matrix = _separation_matrix(n_qubits, state_vector, partition)

    return np.linalg.svd(sep_matrix)

def schmidt_composition(svd_u, svd_v, singular_values, partition):
    """
    Execute the Schmidt composition of a state vector.
    The inverse of the Schmidt decomposition.

    Returns
    -------
    state_vector: list of complex
        A unit vector representing a quantum state.
        Values are amplitudes.
    """

    n_qubits = _to_qubits(svd_u.shape[0]) + _to_qubits(svd_v.shape[1])

    #sep_matrix = svd_u @ np.diag(singular_values) @ svd_v
    sep_matrix = (svd_u * singular_values) @ svd_v

    state_vector = _undo_separation_matrix(n_qubits, sep_matrix, partition)

    return state_vector

def low_rank_approximation(low_rank, svd_u, svd_v, singular_values):
    """
    Low-rank approximation from the SVD.
    """
    rank = singular_values.shape[0] # max rank

    effective_rank = _effective_rank(singular_values)

    if 0 < low_rank < rank or effective_rank < rank:
        if 0 < low_rank < effective_rank:
            effective_rank = low_rank

        # To use isometries, the rank needs to be a power of 2.
        rank = int(2**ceil(log2(effective_rank)))

        svd_u = svd_u[:,:rank]
        svd_v = svd_v[:rank,:]
        singular_values = singular_values[:rank]

    return rank, svd_u, svd_v, singular_values

def _separation_matrix(n_qubits, state_vector, partition):
    new_shape = (2 ** (n_qubits-len(partition)), 2 ** len(partition))

    qubit_shape = tuple([2] * n_qubits)
    # We need to swap qubits from their subsystem2 position to the end of the
    # mode as we expect that we do LSB to be on the left-most side.
    from_move = sorted(partition)
    to_move = (n_qubits - np.arange(1, len(partition) + 1))[::-1]

    sep_matrix = \
        np.moveaxis(np.array(state_vector).reshape(qubit_shape),
                                            from_move, to_move).reshape(new_shape)
    return sep_matrix

def _undo_separation_matrix(n_qubits, sep_matrix, partition):
    new_shape = (2 ** n_qubits, )

    qubit_shape = tuple([2] * n_qubits)

    to_move = sorted(partition)
    from_move = (n_qubits - np.arange(1, len(partition) + 1))[::-1]

    state_vector = \
        np.moveaxis(np.array(sep_matrix).reshape(qubit_shape),
                                            from_move, to_move).reshape(new_shape)
    return state_vector

def _effective_rank(singular_values):
    return sum(j > 10**-7 for j in singular_values)



# Auxiliary functions



def _to_qubits(n_state_vector):
    return int(log2(n_state_vector))

def _default_partition(n_qubits):
    odd = n_qubits % 2
    return list(range(n_qubits//2 + odd))

def _create_quantum_circuit(state_vector, partition):
    n_qubits = _to_qubits(len(state_vector))
    if partition is None:
        partition = _default_partition(n_qubits)

    complement = sorted(set(range(n_qubits)).difference(set(partition)))

    circuit = QuantumCircuit(n_qubits)

    return circuit, partition[::-1], complement[::-1]

def _encode(data, circuit, reg, iso_scheme='ccd', uni_scheme='qsd'):
    """
    Encodes data using the most appropriate method.
    """
    n_qubits = len(reg)
    rank = 0
    if data.shape[1] == 1:
        partition = _default_partition(n_qubits)
        _, svals, _ = schmidt_decomposition(data[:, 0], partition)
        rank = _effective_rank(svals)

    if data.shape[1] == 1 and (n_qubits % 2 == 0 or n_qubits < 4 or rank==1):
        # state preparation
        gate_u = initialize(data[:, 0], isometry_scheme=iso_scheme,
                                            unitary_scheme=uni_scheme)
    elif data.shape[0] > data.shape[1]:
        gate_u = decompose_isometry(data, scheme=iso_scheme)
    else:
        gate_u = decompose_unitary(data, decomposition=uni_scheme)

    # Apply gate U to the register reg
    circuit.compose(gate_u, reg, inplace=True)



# CNOT count



def cnot_count(state_vector, low_rank=0, isometry_scheme='ccd', unitary_scheme='qsd',
                                                        partition=None, method='estimate'):
    """
    Estimate the number of CNOTs to build the state preparation circuit.
    """

    n_qubits = _to_qubits(len(state_vector))
    if n_qubits < 2:
        return 0

    if partition is None:
        partition = _default_partition(n_qubits)

    cnots = 0

    # Schmidt decomposition
    svd_u, singular_values, svd_v = schmidt_decomposition(state_vector, partition)

    rank, svd_u, svd_v, singular_values = \
        low_rank_approximation(low_rank, svd_u, svd_v, singular_values)

    # Schmidt measure of entanglement
    ebits =  _to_qubits(rank)

    # Phase 1.
    if ebits > 0:
        singular_values = singular_values / np.linalg.norm(singular_values)
        cnots += _cnots(singular_values.reshape(rank, 1), isometry_scheme,
                                                          unitary_scheme, method)
    # Phase 2.
    cnots += ebits

    # Phases 3 and 4.
    cnots += _cnots(svd_u, isometry_scheme, unitary_scheme, method)
    cnots += _cnots(svd_v.T, isometry_scheme, unitary_scheme, method)

    return cnots

def _cnots(data, iso_scheme='ccd', uni_scheme='qsd', method='estimate'):
    n_qubits = _to_qubits(data.shape[0])

    rank = 0
    if data.shape[1] == 1:
        partition = _default_partition(n_qubits)
        _, svals, _ = schmidt_decomposition(data[:, 0], partition)
        rank = _effective_rank(svals)

    if data.shape[1] == 1 and (n_qubits % 2 == 0 or n_qubits < 4 or rank==1):
        return cnot_count(data[:, 0], isometry_scheme=iso_scheme,
                                       unitary_scheme=uni_scheme, method=method)
    if data.shape[0] > data.shape[1]:
        return cnots_isometry(data, scheme=iso_scheme, method=method)

    return cnots_unitary(data, decomposition=uni_scheme, method=method)



# Deprecated



@deprecation.deprecated(deprecated_in="0.0.7",
                        details="Use the initialize function instead")
def initialize_original(state_vector):
    """ State preparation using Schmidt decomposition arXiv:1003.5760.
        This function implements the original algorithm as defined in arXiv:1003.5760.
        It is kept here for didactic reasons.
        The ``initialize`` function should preferably be used.

    For instance, to initialize the state a|0> + b|1>
        $ state = [a, b]
        $ circuit = initialize_original(state)

    Parameters
    ----------
    state_vector: list of int
        A unit vector representing a quantum state.
        Values are amplitudes.

    Returns
    -------
    circuit: QuantumCircuit
        QuantumCircuit to initialize the state.
    """

    state = np.copy(state_vector)

    size = len(state)
    n_qubits = _to_qubits(size)

    odd = n_qubits % 2

    state.shape = (int(2**(n_qubits//2)), int(2**(n_qubits//2 + odd)))

    svd_u, singular_values, svd_v = np.linalg.svd(state)

    singular_values = singular_values / np.linalg.norm(singular_values)

    reg_a = QuantumRegister(n_qubits//2 + odd)
    reg_b = QuantumRegister(n_qubits//2)

    circuit = QuantumCircuit(reg_a, reg_b)

    if len(singular_values) > 2:
        circ = initialize_original(singular_values)
    else:
        circ = mottonen(singular_values)
    circuit.compose(circ, reg_b, inplace=True)

    for k in range(int(n_qubits//2)):
        circuit.cx(reg_b[k], reg_a[k])

    gate_u = decompose_unitary(svd_u, 'qsd')
    gate_v = decompose_unitary(svd_v.T, 'qsd')

    circuit.compose(gate_u, reg_b, inplace=True) # apply gate U to the first register
    circuit.compose(gate_v, reg_a, inplace=True) # apply gate V to the second register

    return circuit



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the baa_schmidt.py module.
"""

from unittest import TestCase
import numpy as np
from qiskit import ClassicalRegister, execute
from qiskit.providers.aer.backends import AerSimulator
from qclib.util import get_state
from qclib.state_preparation.baa_schmidt import initialize

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

class TestBaaSchmidt(TestCase):
    @staticmethod
    def fidelity(state1, state2):
        bra = np.conj(state1)
        ket = state2

        return np.power(np.abs(bra.dot(ket)), 2)

    @staticmethod
    def get_counts(circuit):
        n_qubits = circuit.num_qubits
        classical_reg = ClassicalRegister(n_qubits)
        circuit.add_register(classical_reg)
        circuit.measure(list(range(n_qubits)), classical_reg)

        backend = AerSimulator()
        counts = execute(circuit, backend, shots=8192).result().get_counts()

        counts_with_zeros = {}
        for i in range(2**n_qubits):
            pattern = '{:0{}b}'.format(i, n_qubits)
            if pattern in counts:
                counts_with_zeros[pattern] = counts[pattern]
            else:
                counts_with_zeros[pattern] = 0.0

        sum_values = sum(counts.values())
        return [ value/sum_values for (key, value) in counts_with_zeros.items() ]

    def _test_initialize_loss(self, fidelity_loss, state_vector=None,
                                    n_qubits=5, strategy='brute_force', use_low_rank=False):
        if state_vector is None:
            state_vector = np.random.rand(2**n_qubits) + np.random.rand(2**n_qubits) * 1j
            state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector, max_fidelity_loss=fidelity_loss, strategy=strategy,
                                                                    use_low_rank=use_low_rank)

        state = get_state(circuit)

        fidelity = TestBaaSchmidt.fidelity(state_vector, state)

        self.assertTrue(round(fidelity,2)>=round(1-fidelity_loss,2)*0.99)

    def test_initialize_loss_brute_force(self):
        for loss in range(5, 15):
            self._test_initialize_loss(loss/100, n_qubits=5, strategy='brute_force')

    def test_initialize_loss_brute_force_low_rank(self):
        for loss in range(5, 15):
            self._test_initialize_loss(loss/100, n_qubits=5, strategy='brute_force',
                                                                    use_low_rank=True)

    def test_initialize_loss_greedy(self):
        for loss in range(5, 15):
            self._test_initialize_loss(loss/100, n_qubits=5, strategy='greedy')

    def test_initialize_loss_greedy_low_rank(self):
        for loss in range(5, 15):
            self._test_initialize_loss(loss/100, n_qubits=5, strategy='greedy',
                                                                    use_low_rank=True)

    def test_initialize_loss_fixed_n3(self):
        state_vector = [-0.33*1j,0,-0.44-0.44*1j,0.24+0.23*1j,0,0,0,0.62-0.01*1j]
        state_vector = state_vector/np.linalg.norm(state_vector)
        for loss in [0.1, 0.28, 0.9]:
            self._test_initialize_loss(loss, state_vector=state_vector, strategy='brute_force')
            self._test_initialize_loss(loss, state_vector=state_vector, strategy='greedy')

    def test_initialize_loss_fixed_n4(self):
        state_vector = [0.04214906+0.25870366j, 0.18263984+0.05596082j, 0.17202687+0.1843925j ,
                        0.24972444+0.04666321j, 0.03311006+0.28233458j, 0.26680588+0.22211721j,
                        0.07205056+0.04556719j, 0.27982261+0.01626855j, 0.22908475+0.25461504j,
                        0.14290823+0.2425394j , 0.14213592+0.08282699j, 0.0068727 +0.03378424j,
                        0.2016483 +0.298073j  , 0.07520782+0.0639856j , 0.01026576+0.07669651j,
                        0.31755857+0.09279232j]
        state_vector = state_vector/np.linalg.norm(state_vector)
        for loss in [0.1, 0.15, 0.18, 0.2]:
            self._test_initialize_loss(loss, state_vector=state_vector, strategy='brute_force')
            self._test_initialize_loss(loss, state_vector=state_vector, strategy='greedy')

    def test_initialize_loss_fixed_n5(self):
        state_vector = [0.17777766+0.10171662j, 0.19896424+0.10670792j, 0.07982054+0.19653055j,
                        0.18155708+0.05746777j, 0.04259147+0.17093567j, 0.21551328+0.08246133j,
                        0.09549255+0.1117806j , 0.20562749+0.12218064j, 0.16191832+0.01653411j,
                        0.12255337+0.14109365j, 0.20090638+0.11119666j, 0.19851901+0.04543331j,
                        0.06842539+0.16671467j, 0.03209685+0.16839388j, 0.01707365+0.20060943j,
                        0.03853768+0.08183117j, 0.00073591+0.10084589j, 0.09524694+0.18785593j,
                        0.06005853+0.06977443j, 0.01553849+0.05363906j, 0.10294799+0.12558734j,
                        0.20142903+0.06801796j, 0.05282011+0.20879126j, 0.11257846+0.20746226j,
                        0.17737416+0.03461382j, 0.01689154+0.06600272j, 0.06428148+0.06199636j,
                        0.1163249 +0.160533j  , 0.14177201+0.10456823j, 0.03156739+0.04567818j,
                        0.02078566+0.02023752j, 0.18967059+0.03469463j]
        state_vector = state_vector/np.linalg.norm(state_vector)
        for loss in [0.1, 0.12, 0.14, 0.16, 0.2]:
            self._test_initialize_loss(loss, state_vector=state_vector, strategy='brute_force')
            self._test_initialize_loss(loss, state_vector=state_vector, strategy='greedy')

    def test_initialize_no_loss(self):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state))

    def test_initialize_ame(self):
        """ Test initialization of a absolutely maximally entangled state"""
        state_vector = [1, 1, 1, 1,1,-1,-1, 1, 1,-1,-1, 1, 1, 1,1,1,
                        1, 1,-1,-1,1,-1, 1,-1,-1, 1,-1, 1,-1,-1,1,1]
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state))

    def test_measurement_no_loss(self):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector)

        state = TestBaaSchmidt.get_counts(circuit)

        self.assertTrue(np.allclose( np.power(np.abs(state_vector),2), state,
                        rtol=1e-01, atol=0.005))

    def test_compare_strategies(self):
        fidelities1 = []
        fidelities2 = []
        for n_qubits in range(3,7):
            state_vector = np.random.rand(2**n_qubits) + np.random.rand(2**n_qubits) * 1j
            state_vector = state_vector / np.linalg.norm(state_vector)
            for loss in range(10, 20):
                circuit = initialize(state_vector, max_fidelity_loss=loss/100,
                                                        strategy='brute_force')
                state = get_state(circuit)
                fidelity1 = TestBaaSchmidt.fidelity(state_vector, state)

                circuit = initialize(state_vector, max_fidelity_loss=loss/100,
                                                        strategy='greedy')
                state = get_state(circuit)
                fidelity2 = TestBaaSchmidt.fidelity(state_vector, state)

                fidelities1.append(fidelity1)
                fidelities2.append(fidelity2)

        self.assertTrue(np.allclose(fidelities1, fidelities2, rtol=0.125, atol=0.0))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Majority gate """

from itertools import combinations
import numpy as np

def operate(circuit, controls, target):
    """ Apply a majority gate """

    size_controls = len(controls)
    log_n = int(np.floor(np.log2(size_controls)))

    n_min = int(np.ceil(size_controls/2))
    n_max = 2**log_n

    n_controls = []
    n_controls.append(n_min)

    if n_min != n_max:
        if n_min % 2 != 0:
            n_controls.extend(range(n_min+1, n_max))

        n_controls.append(n_max)

    for k in n_controls:
        comb = combinations(controls, k)
        for j in comb:
            circuit.mcx([*j], target)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    Divide-and-conquer state preparation
    https://arxiv.org/abs/2108.10182
"""

import numpy as np
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister

from qclib.state_preparation.util.state_tree_preparation import Amplitude, state_decomposition

from qclib.state_preparation.util.angle_tree_preparation import \
    create_angles_tree

from qclib.state_preparation.util.tree_register import add_register
from qclib.state_preparation.util.tree_walk import bottom_up

def initialize(state):
    """
        https://arxiv.org/abs/2108.10182
    """
    n_qubits = int(np.log2(len(state)))
    data = [Amplitude(i, a) for i, a in enumerate(state)]

    state_tree = state_decomposition(n_qubits, data)
    angle_tree = create_angles_tree(state_tree)

    circuit = QuantumCircuit()
    add_register(circuit, angle_tree, n_qubits-1)

    bottom_up(angle_tree, circuit, n_qubits)

    return circuit






class BinTree:
    """
        State preparation using DCSP https://www.nature.com/articles/s41598-021-85474-1.
        This class implements the original algorithm as defined in the paper.
        It is kept here for didactic reasons.
        The ``initialize`` function should preferably be used.
    """
    size = None
    values = None

    def __init__(self, values):
        self.size = len(values)
        self.values = values


    @staticmethod
    def parent(key):
        """ returns parent node index"""
        return int((key-0.5)/2)

    @staticmethod
    def left(key):
        """ returns left child index"""
        return int(2 * key + 1)

    @staticmethod
    def right(key):
        """ returns right child index"""
        return int(2 * key + 2)

    @staticmethod
    def root():
        """ returns root index """
        return 0

    def __getitem__(self, key):
        return self.values[key]


class Encoding:
    """
    Encode information into qubits
    """
    qcircuit = None
    quantum_data = None
    classical_data = None
    num_qubits = None
    tree = None
    output_qubits = []

    def __init__(self, input_vector, encode_type='amplitude_encoding'):
        if encode_type == 'amplitude_encoding':
            self.amplitude_encoding(input_vector)
        if encode_type == 'qubit_encoding':
            self.qubit_encoding(input_vector)
        if encode_type == 'dc_amplitude_encoding':
            self.dc_amplitude_encoding(input_vector)
        if encode_type == 'basis_encoding':
            self.basis_encoding(input_vector)

    def basis_encoding(self, input_vector, n_classical=1):
        """
        encoding a binary string x in a basis state |x>
        """
        self.num_qubits = int(len(input_vector))
        self.quantum_data = QuantumRegister(self.num_qubits)
        self.classical_data = ClassicalRegister(n_classical)
        self.qcircuit = QuantumCircuit(self.quantum_data, self.classical_data)
        for k, _ in enumerate(input_vector):
            if input_vector[k] == 1:
                self.qcircuit.x(self.quantum_data[k])


    def qubit_encoding(self, input_vector, n_classical=1):
        """
        encoding a binary string x as
        """
        input_pattern = QuantumRegister(len(input_vector))
        classical_register = ClassicalRegister(n_classical)
        self.qcircuit = QuantumCircuit(input_pattern, classical_register)
        for k, _ in enumerate(input_vector):
            self.qcircuit.ry(input_vector[k], input_pattern[k])

    @staticmethod
    def _recursive_compute_beta(input_vector, betas):
        if len(input_vector) > 1:
            new_x = []
            beta = []
            for k in range(0, len(input_vector), 2):
                norm = np.sqrt(input_vector[k] ** 2 + input_vector[k + 1] ** 2)
                new_x.append(norm)
                if norm == 0:
                    beta.append(0)
                else:
                    if input_vector[k] < 0:
                        ## testing
                        beta.append(2 * np.pi - 2 * np.arcsin(input_vector[k + 1] / norm))
                    else:
                        beta.append(2 * np.arcsin(input_vector[k + 1] / norm))
            Encoding._recursive_compute_beta(new_x, betas)
            betas.append(beta)

    @staticmethod
    def _index(k, circuit, control_qubits, numberof_controls):
        binary_index = '{:0{}b}'.format(k, numberof_controls)
        for j, qbit in enumerate(control_qubits):
            if binary_index[j] == '1':
                circuit.x(qbit)


    def amplitude_encoding(self, input_vector):
        """
        load real vector x to the amplitude of a quantum state
        """
        self.num_qubits = int(np.log2(len(input_vector)))
        self.quantum_data = QuantumRegister(self.num_qubits)
        self.qcircuit = QuantumCircuit(self.quantum_data)
        newx = np.copy(input_vector)
        betas = []
        Encoding._recursive_compute_beta(newx, betas)
        self._generate_circuit(betas, self.qcircuit, self.quantum_data)

    def dc_amplitude_encoding(self, input_vector):
        """
        Divide and conquer amplitude encoding
        """
        self.num_qubits = int(len(input_vector))-1
        self.quantum_data = QuantumRegister(self.num_qubits)
        self.qcircuit = QuantumCircuit(self.quantum_data)
        newx = np.copy(input_vector)
        betas = []
        Encoding._recursive_compute_beta(newx, betas)
        self._dc_generate_circuit(betas, self.qcircuit, self.quantum_data)

    def _dc_generate_circuit(self, betas, qcircuit, quantum_input):

        k = 0
        linear_angles = []
        for angles in betas:
            linear_angles = linear_angles + angles
            for angle in angles:
                qcircuit.ry(angle, quantum_input[k])
                k += 1

        self.tree = BinTree(quantum_input)
        my_tree = self.tree

        last = my_tree.size - 1
        actual = my_tree.parent(last)
        level = my_tree.parent(last)
        while actual >= 0:
            left_index = my_tree.left(actual)
            right_index = my_tree.right(actual)
            while right_index <= last:

                qcircuit.cswap(my_tree[actual], my_tree[left_index], my_tree[right_index])

                left_index = my_tree.left(left_index)
                right_index = my_tree.left(right_index)
            actual -= 1
            if level != my_tree.parent(actual):
                level -= 1

        # set output qubits
        next_index = 0
        while next_index < my_tree.size:
            self.output_qubits.append(next_index)
            next_index = my_tree.left(next_index)

    def _generate_circuit(self, betas, qcircuit, quantum_input):
        numberof_controls = 0  # number of controls
        control_bits = []
        for angles in betas:
            if numberof_controls == 0:
                qcircuit.ry(angles[0], quantum_input[self.num_qubits-1])
                numberof_controls += 1
                control_bits.append(quantum_input[self.num_qubits-1])
            else:
                for k, angle in enumerate(reversed(angles)):
                    Encoding._index(k, qcircuit, control_bits, numberof_controls)

                    qcircuit.mcry(angle,
                                  control_bits,
                                  quantum_input[self.num_qubits - 1 - numberof_controls],
                                  None,
                                  mode='noancilla')

                    Encoding._index(k, qcircuit, control_bits, numberof_controls)
                control_bits.append(quantum_input[self.num_qubits - 1 - numberof_controls])
                numberof_controls += 1



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the baa.py module.
"""
import sys

from test.test_baa_schmidt import TestBaaSchmidt
import datetime
import os
from multiprocessing import Pool
from unittest import TestCase

import numpy as np
import qiskit
from qiskit import QiskitError
from qiskit.circuit.random import random_circuit

from qclib.entanglement import geometric_entanglement, \
    meyer_wallach_entanglement
from qclib.state_preparation.baa_schmidt import initialize
from qclib.state_preparation.schmidt import cnot_count as schmidt_cnots
from qclib.util import get_state

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring


use_parallel = os.getenv('QLIB_TEST_PARALLEL', 'False') == 'True'


def get_vector(e_lower: float, e_upper: float, num_qubits: int,
               start_depth_multiplier=1, measure='meyer_wallach'):
    entanglement = -1.0

    if isinstance(start_depth_multiplier, int):
        multiplier = start_depth_multiplier
    elif isinstance(start_depth_multiplier, (tuple, list, np.ndarray)):
        assert len(start_depth_multiplier) > 1
        multiplier = np.random.randint(start_depth_multiplier[0], start_depth_multiplier[1])
    else:
        raise ValueError("start_depth_multiplier must be either an int or an array-like of int.")

    iteration = 0
    entanglements = []
    vector = np.ndarray(shape=(0,))
    while e_lower > entanglement or entanglement > e_upper:
        np.random.seed()
        q_circuit: qiskit.QuantumCircuit = random_circuit(num_qubits, multiplier * num_qubits)
        vector = get_state(q_circuit)
        if measure == 'geometric':
            entanglement = geometric_entanglement(vector)
        elif measure == 'meyer_wallach':
            entanglement = meyer_wallach_entanglement(vector)
        else:
            raise ValueError(f'Entanglement Measure {measure} unknown.')
        iteration += 1
        if iteration > 100:
            multiplier += 1
            iteration = 0
            if not use_parallel:
                print(
                    f'{multiplier} ({np.min(entanglements):.4f}-{np.max(entanglements):.4f})',
                    end='\n', flush=True
                )
            entanglements = []
        else:
            entanglements.append(entanglement)
            if not use_parallel:
                print('.', end='', flush=True)
    if not use_parallel:
        print(
            f'Final {multiplier} ({np.min(entanglements):.4f}-{np.max(entanglements):.4f})',
            end='\n', flush=True
        )
    return vector, entanglement, multiplier * num_qubits


def initialize_loss(fidelity_loss, state_vector=None, n_qubits=5, strategy='brute_force',
                    use_low_rank=False):

    if state_vector is None:
        state_vector = np.random.rand(2**n_qubits) + np.random.rand(2**n_qubits) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

    circuit, node = initialize(
        state_vector, max_fidelity_loss=fidelity_loss, strategy=strategy, use_low_rank=use_low_rank,
        return_node=True
    )
    state = get_state(circuit)
    fidelity = TestBaaSchmidt.fidelity(state_vector, state)

    try:
        basis_circuit = qiskit.transpile(
            circuit, basis_gates=['rx', 'ry', 'rz', 'cx'], optimization_level=0
        )
        cnots = len([d[0] for d in basis_circuit.data if d[0].name == 'cx'])
        depth = basis_circuit.depth()
    except QiskitError as ex:
        print(ex)
        return -1, -1, -1, node

    return cnots, depth, round(1 - fidelity, 4), node


def execute_experiment(exp_idx,  num_qubits, entanglement_bounds,
                       max_fidelity_losses, return_state=False,
                       start_depth_multiplier=1):
    print(f"Starting {exp_idx,  num_qubits, entanglement_bounds, max_fidelity_losses}")

    # State Generation
    state_vector, entganglement, depth = get_vector(
        *entanglement_bounds, num_qubits, start_depth_multiplier=start_depth_multiplier,
        measure='geometric'
    )
    mw_entanglement = meyer_wallach_entanglement(state_vector)
    geo_entanglement = geometric_entanglement(state_vector)
    cnots = schmidt_cnots(state_vector)
    if not use_parallel:
        print(
            f"Found state for entanglement bounds {entganglement} in {entanglement_bounds}. "
            f"State preparation needs {cnots}."
        )

    # Benchmark against real Algorithm
    real_cnots_benchmark, real_depth_benchmark, real_fidelity_loss_benchmark, _ = initialize_loss(
        state_vector=state_vector, fidelity_loss=0.0, use_low_rank=False
    )
    data_result = []
    for max_fidelity_loss in max_fidelity_losses:
        for use_low_rank in [False, True]:
            for strategy in ['brute_force', 'greedy']:
                if not use_parallel:
                    print(
                        f"[{max_fidelity_loss}] {strategy.upper()} "
                        f"{'With' if use_low_rank else 'No'} Low Rank Processing....", end=''
                    )
                start_time = datetime.datetime.now()
                real_cnots, real_depth, real_fidelity_loss, node = initialize_loss(
                    state_vector=state_vector, fidelity_loss=max_fidelity_loss,
                    use_low_rank=use_low_rank, strategy=strategy
                )
                end_time = datetime.datetime.now()
                duration = (end_time - start_time).total_seconds()
                vector_data = list(
                    zip(node.ranks, [list(v.shape) for v in node.vectors])
                )
                data = [
                    exp_idx, use_low_rank, strategy, num_qubits, depth, cnots, geo_entanglement,
                    mw_entanglement, max_fidelity_loss, node.total_saved_cnots,
                    node.total_fidelity_loss, vector_data, real_cnots, real_cnots_benchmark,
                    real_depth, real_depth_benchmark, real_fidelity_loss,
                    real_fidelity_loss_benchmark, duration
                ]
                if not use_parallel:
                    print(f"in {duration} secs")
                data_result.append(data)

    # Experiment transcription
    data_result = np.asarray(data_result, dtype=object)
    print(f"Done {exp_idx,  num_qubits, entanglement_bounds, max_fidelity_losses}")
    if return_state:
        return data_result, state_vector
    return data_result


class TestBaa(TestCase):
    columns = [
        'id', 'with_low_rank', 'strategy', 'num_qubits', 'depth', 'cnots', 'entganglement',
        'entganglement (MW)', 'max_fidelity_loss', 'total_saved_cnots', 'total_fidelity_loss',
        'data', 'real_cnots', 'real_cnots_benchmark', 'real_depth', 'real_depth_no_approx',
        'real_fidelity_loss', 'real_fidelity_loss_benchmark', 'duration',
        'predicted_cnots', 'benchmark_fidelity_loss_pass', 'approximation_calculation_pass',
        'real_approximation_calculation_pass'
    ]

    def test(self):
        num_qubits = 7
        entanglement_bounds = (0.4, 1.0)
        max_fidelity_loss = np.linspace(0.1, 1.0, 4)
        number_of_experiments = 5

        data = [
            (i, num_qubits, entanglement_bounds, max_fidelity_loss)
            for i in range(number_of_experiments)
        ]
        if use_parallel:
            with Pool() as pool:
                result = pool.starmap(execute_experiment, data)
        else:
            result = [execute_experiment(*d) for d in data]

        test_data = np.concatenate(result)
        # Predicted CNOTs
        test_data = np.hstack([
            test_data,
            (test_data[:, self.columns.index('cnots')]
             - test_data[:, self.columns.index('total_saved_cnots')]).reshape(-1, 1)
        ])

        # Tests:
        # Calculation Tests
        # The benchmark will not change the state at all, so it must be
        # essentially zero
        test_data = np.hstack([
            test_data,
            (np.abs(test_data[:, self.columns.index('real_fidelity_loss_benchmark')]) < 1e-6
             ).reshape(-1, 1)
        ])
        # The expected / predicted fidelity loss must be less or equal to the
        # max fidelity loss
        test_data = np.hstack([
            test_data,
            (test_data[:, self.columns.index('max_fidelity_loss')]
             >= test_data[:, self.columns.index('total_fidelity_loss')]
             ).reshape(-1, 1)
        ])

        # The real Tests
        # The real measured fidelity measure must be less or equal to the
        # configured mx fidelity loss
        test_data = np.hstack([
            test_data,
            (test_data[:, self.columns.index('real_fidelity_loss')]
             - test_data[:, self.columns.index('max_fidelity_loss')] < 0.1
             ).reshape(-1, 1)
        ])

        # Attention: this is a set of tests that tests how well the CNOT-estimation works
        # However, we do not test this here. As a result, also the prediciton of CNOTs cannot
        # reliably be tested.
        # START: COMMENTED OUT AND LEFT FOR DOCUMENTATION REASONS
        # # The predicted maximum CNOT gates and the no-approximation count
        # # should be within 10%
        # test_data['cnot_prediction_calculation_pass'] = (
        #         np.abs(test_data['real_cnots_benchmark']
        #         - test_data['cnots']) < 0.2 * test_data['cnots']
        # )
        # # The predicted CNOT gates should be within an error margin of 10%
        # test_data['saved_cnots_calculation_pass'] = (
        #         np.abs(test_data['predicted_cnots']
        #         - test_data['real_cnots']) <= 0.2 * (test_data['predicted_cnots'])
        # )
        # END: COMMENTED OUT AND LEFT FOR DOCUMENTATION REASONS

        test_passed = True
        total_experiments = number_of_experiments * max_fidelity_loss.shape[0] * 4
        fails_to_still_pass = int(np.ceil(total_experiments * 0.01))
        benchmark_fidelity_loss_fail_count = (
                test_data.shape[0]
                - test_data[
                    test_data[:, self.columns.index('benchmark_fidelity_loss_pass')].astype(bool)
                ].shape[0]
        )
        approximation_calculation_fail_count = (
                test_data.shape[0]
                - test_data[
                    test_data[:, self.columns.index('approximation_calculation_pass')].astype(bool)
                ].shape[0]
        )
        real_approximation_calculation_fail_count = (
                test_data.shape[0] - test_data[
                test_data[:, self.columns.index('real_approximation_calculation_pass')].astype(bool)
            ].shape[0]
        )
        if benchmark_fidelity_loss_fail_count > 0:
            print(f"[WARNING] NOT ALL benchmark_fidelity_loss_pass are true ({benchmark_fidelity_loss_fail_count})",
                  file=sys.stderr)
        if benchmark_fidelity_loss_fail_count > fails_to_still_pass:
            print("[FAIL] benchmark_fidelity_loss_pass must be true", file=sys.stderr)
            test_passed = False
        if approximation_calculation_fail_count > 0:
            print(f"[WARNING] NOT ALL approximation_calculation_pass are true ({approximation_calculation_fail_count})",
                  file=sys.stderr)
        if approximation_calculation_fail_count > fails_to_still_pass:
            print("[FAIL] approximation_calculation_pass must be true", file=sys.stderr)
            test_passed = False
        if real_approximation_calculation_fail_count > 0:
            print(f"[WARNING] NOT ALL real_approximation_calculation_pass are true "
                  f"({real_approximation_calculation_fail_count})", file=sys.stderr)
        if real_approximation_calculation_fail_count > fails_to_still_pass:
            print("[FAIL] real_approximation_calculation_pass must be true", file=sys.stderr)
            test_passed = False

        import importlib.util
        pandas_loader = importlib.util.find_spec('pandas')
        if pandas_loader is None:
            with np.printoptions(precision=2, linewidth=1000, suppress=True, threshold=sys.maxsize, floatmode='fixed'):
                print(test_data)
        else:
            import pandas as pd
            print(pd.DataFrame(test_data, columns=self.columns).to_string())

        self.assertTrue(test_passed, 'The tests should all pass.')

    def test_no_ops(self):
        # The Test is based on randomly generated states. They all should
        # create the correct fidelity (1.0). After 10 attempts we may well
        # find one that fails. If it doesn't it is probably okay.
        for lower_bound_int in range(10):
            state_vector, _, depth = get_vector(
                lower_bound_int/10, 1.0, 8, 1, measure='geometric'
            )
            cnots, depth, fidelity_loss, _ = initialize_loss(0.0, state_vector)
            if cnots == depth == fidelity_loss == -1:
                continue
            self.assertAlmostEqual(0.0, fidelity_loss, 4)

    def test_node_state_vector(self):
        """
        The method Node.state_vector() is an important function for analytics, but it was wrong. The ordering
        of the qubits must be taken into account. One example, visible to the human eye, was a probability
        distribution, the log-normal. For a max-fidelity-loss of 0.09, the ordering is computed with qubits
        [(1,), (3,), (0, 2, 4, 5, 6)] and partitions [None, None, (1, 4)].
        The resulting state has an actual fidelity loss of about 0.0859261062108.
        The fix is to take the ordering into account. Here we test that for this example, the
        fidelity is actually correct.

        """

        import scipy.stats as stats
        from qclib.state_preparation.util import baa

        # geometric measure = 0.11600417225836746
        state = [0.07790067, 0.12411293, 0.10890448, 0.09848761, 0.05027826, 0.05027438,
                 0.03067742, 0.11846638, 0.09868947, 0.10711022, 0.01826258, 0.12535963,
                 0.11613662, 0.05865527, 0.05427737, 0.05451262, 0.07021045, 0.09220849,
                 0.08365776, 0.06869252, 0.09956702, 0.04754114, 0.0688004,  0.07704547,
                 0.08596225, 0.11279128, 0.05687908, 0.09127936, 0.09797266, 0.02743385,
                 0.09921588, 0.05256358, 0.03246542, 0.1239935,  0.12508287, 0.11444701,
                 0.07025331, 0.03978115, 0.10529168, 0.08444882, 0.04446722, 0.08957199,
                 0.02360472, 0.12138094, 0.06475262, 0.10360775, 0.07106703, 0.09179565,
                 0.09411756, 0.05472767, 0.12533861, 0.1120676,  0.12337869, 0.12040975,
                 0.0984252,  0.12221594, 0.03786564, 0.05635093, 0.02707025, 0.07260295,
                 0.07935725, 0.06630651, 0.11587787, 0.07602843, 0.06746749, 0.09377139,
                 0.04778426, 0.11400727, 0.03475503, 0.12645201, 0.11185863, 0.05674245,
                 0.00945899, 0.11494597, 0.10701826, 0.10868207, 0.11178804, 0.03463689,
                 0.07621068, 0.04332871, 0.11825607, 0.10049395, 0.07322158, 0.03209064,
                 0.0709839,  0.07258655, 0.10872672, 0.10163697, 0.11989633, 0.08747055,
                 0.04401971, 0.10750071, 0.11102557, 0.09536318, 0.11176607, 0.08944697,
                 0.09203053, 0.0832302,  0.02029422, 0.04181051, 0.02256621, 0.10154549,
                 0.07136789, 0.0907753,  0.12126382, 0.06355451, 0.08154299, 0.110643,
                 0.06088612, 0.03531675, 0.06851802, 0.05110969, 0.12273343, 0.11442741,
                 0.10130534, 0.11882721, 0.11411204, 0.05498105, 0.12025703, 0.09348119,
                 0.11437924, 0.12049476, 0.07178074, 0.04222706, 0.06077118, 0.08318802,
                 0.11512578, 0.1180934 ]

        node = baa.adaptive_approximation(state, use_low_rank=True, max_fidelity_loss=0.09, strategy='brute_force')
        fidelity = np.vdot(state, node.state_vector())**2

        self.assertAlmostEqual(node.total_fidelity_loss, 1 - fidelity, places=4)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
https://arxiv.org/abs/2108.10182
"""

import qiskit
from qclib.state_preparation.util.tree_utils import children


def output(angle_tree, output_qubits):
    """ Define output qubits"""
    if angle_tree:
        output_qubits.insert(0, angle_tree.qubit) # qiskit little-endian
        if angle_tree.left:
            output(angle_tree.left, output_qubits)
        else:
            output(angle_tree.right, output_qubits)

def _add_register(angle_tree, qubits, start_level):
    if angle_tree:
        angle_tree.qubit = qubits.pop(0)
        if angle_tree.level < start_level:
            _add_register(angle_tree.left, qubits, start_level)
            _add_register(angle_tree.right, qubits, start_level)
        else:
            if angle_tree.left:
                _add_register(angle_tree.left, qubits, start_level)
            else:
                _add_register(angle_tree.right, qubits, start_level)

def add_register(circuit, angle_tree, start_level):
    """
    Organize qubit registers, grouping by "output" and "ancilla" types.
    """

    level = 0
    level_nodes = []
    nodes = [angle_tree]
    while len(nodes) > 0: # count nodes per level
        level_nodes.append( len(nodes) )
        nodes = children(nodes)
        level += 1

    noutput = level # one output qubits per level
    nqubits = sum(level_nodes[:start_level]) # bottom-up qubits

    # top-down qubits: (number of sub-states) * (number of qubits per sub-state)
    nqubits += level_nodes[start_level] * (noutput-start_level)

    nancilla = nqubits - noutput

    output_register = qiskit.QuantumRegister(noutput, name='output')
    circuit.add_register(output_register)
    qubits = [*output_register[::-1]]

    if nancilla > 0:
        ancilla_register = qiskit.QuantumRegister(nancilla, name='ancilla')
        circuit.add_register(ancilla_register)
        qubits.extend([*ancilla_register[::-1]])

    _add_register(angle_tree, qubits, start_level)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test Probabilistic Quantum Memory"""

from unittest import TestCase
import numpy as np
from qiskit import ClassicalRegister, QuantumRegister, QuantumCircuit
from qclib.memory.pqm import initialize as pqm
from qclib.state_preparation.schmidt import initialize
from qclib.util import get_counts

class TestPQM(TestCase):
    """ Testing qclib.memory.pqm"""
    @staticmethod
    def _run_pqm(is_classical_pattern):
        """ run PQM with classical or quantum input"""
        # pqm memory data
        data = [0, 1, 0, 0]
        data = data / np.linalg.norm(data)

        # initialize quantum registers
        memory = QuantumRegister(2, 'm')
        aux = QuantumRegister(1, 'c')
        output = ClassicalRegister(1)
        circ = QuantumCircuit(memory, aux, output)

        # initialize data
        init_gate = initialize(data)
        circ.append(init_gate, memory)

        # initialize input pattern
        bin_input = [1, 0]

        if is_classical_pattern:
            # run pqm recovery algorithm
            pqm(circ, bin_input, memory, aux, is_classical_pattern=True)
        else:
            q_bin_input = QuantumRegister(2)
            circ.add_register(q_bin_input)

            # Pattern basis encoding
            for k, bit in enumerate(bin_input):
                if bit == 1:
                    circ.x(q_bin_input[k])
            # run pqm recovery algorithm
            pqm(circ, q_bin_input, memory, aux, is_classical_pattern=False)

        # measure output and verify results
        circ.measure(aux, output) # pylint: disable=maybe-no-member
        counts = get_counts(circ)

        return counts

    def test_classical_input(self):
        """ Testing PQM with classical input """
        # classical input pattern
        counts = TestPQM._run_pqm(True)

        self.assertTrue(counts['0'] / 1024 == 1)

    def test_quantum_input(self):
        """ Testing PQM with quantum input """
        # quantum input pattern
        counts = TestPQM._run_pqm(False)

        self.assertTrue(counts['0'] / 1024 == 1)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
        https://arxiv.org/abs/quant-ph/0407010
        https://arxiv.org/abs/2108.10182
"""

import numpy as np
from qiskit import QuantumCircuit

from qclib.state_preparation.util.state_tree_preparation import Amplitude, state_decomposition

from qclib.state_preparation.util.angle_tree_preparation import \
    create_angles_tree

from qclib.state_preparation.util.tree_register import add_register
from qclib.state_preparation.util.tree_walk import top_down


def initialize(state, global_phase=True):
    """
        https://arxiv.org/abs/quant-ph/0407010
        https://arxiv.org/abs/2108.10182
    """
    n_qubits = int(np.log2(len(state)))
    data = [Amplitude(i, a) for i, a in enumerate(state)]

    state_tree = state_decomposition(n_qubits, data)
    angle_tree = create_angles_tree(state_tree)

    circuit = QuantumCircuit()
    add_register(circuit, angle_tree, 0)

    top_down(angle_tree, circuit, 0)
    if global_phase:
        # equivalent to unitary(I * exp(1j*sum(np.angle(state))/len(state)))
        circuit.global_phase += sum(np.angle(state))/len(state)

    return circuit



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" sparse state preparation """

import numpy as np
import qiskit
from qclib.state_preparation.schmidt import initialize as dense_init

# pylint: disable=maybe-no-member

class PivotStatePreparation:
    """ Pivot State Preparation """
    def __init__(self, state, aux=False):
        self.aux = aux
        self.n_qubits = len(list(state.keys())[0])
        self.n_qubits = int(self.n_qubits)

        non_zero = len(state)
        self.target_size = np.log2(non_zero)
        self.target_size = np.ceil(self.target_size)
        self.target_size = int(self.target_size)

        self.memory = qiskit.QuantumRegister(self.n_qubits, name='q')
        if self.aux:
            remain = list(range(self.n_qubits - self.target_size, self.n_qubits))
            n_anci = len(remain)

            anc = qiskit.QuantumRegister(n_anci - 1, name='anc')
            pivot_circuit = qiskit.QuantumCircuit(anc, self.memory)

        else:
            pivot_circuit = qiskit.QuantumCircuit(self.memory)

        self.next_state = state.copy()
        index_nonzero = self._get_index_nz(self.n_qubits - self.target_size)

        while index_nonzero is not None:
            self.index_zero = self._get_index_zero(non_zero, self.next_state)
            circ = self.pivoting(index_nonzero, self.target_size, aux)

            pivot_circuit.compose(circ, pivot_circuit.qubits, inplace=True)

            index_nonzero = self._get_index_nz(self.n_qubits - self.target_size)

        dense_state = np.zeros(2 ** self.target_size, dtype=complex)
        for key, value in self.next_state.items():
            dense_state[int(key, 2)] = value

        if non_zero <= 2:
            initialize_circ = qiskit.QuantumCircuit(1)
            initialize_circ.initialize(dense_state)
        else:
            initialize_circ = dense_init(dense_state)

        if self.aux:
            self.sp_circuit = qiskit.QuantumCircuit(anc, self.memory)
            nun_aux = n_anci - 1
            self.sp_circuit.compose(
                initialize_circ,
                self.sp_circuit.qubits[nun_aux:nun_aux + self.target_size],
                inplace=True)

            self.sp_circuit.barrier()
            self.sp_circuit.compose(pivot_circuit.reverse_bits().reverse_ops(), inplace=True)
        else:
            self.sp_circuit = qiskit.QuantumCircuit(self.n_qubits)
            self.sp_circuit.compose(
                initialize_circ,
                self.sp_circuit.qubits[:self.target_size], inplace=True)

            self.sp_circuit.compose(
                pivot_circuit.reverse_bits().reverse_ops(),
                self.sp_circuit.qubits,
                inplace=True)

    def _next_state(self, ctrl_state, index_differ, remain, target_cx):
        tab = {'0': '1', '1': '0'}
        new_state = {}
        for index, _ in self.next_state.items():

            if index[index_differ] == ctrl_state:

                n_index = ''
                for k, _ in enumerate(index):
                    if k in target_cx:
                        n_index = n_index + tab[index[k]]
                    else:
                        n_index = n_index + index[k]

            else:
                n_index = index

            if n_index[remain[0]:] == self.index_zero[remain[0]:]:
                n_index = n_index[:index_differ] + \
                          tab[index[index_differ]] + n_index[index_differ + 1:]

            new_state[n_index] = self.next_state[index]
        return new_state

    def pivoting(self, index_nonzero, target_size, aux=False):
        """ pivot amplitudes of index_nonzero and self.index_zero"""

        n_qubits = len(self.index_zero)
        target = list(range(n_qubits - target_size))
        remain = list(range(n_qubits - target_size, n_qubits))

        memory = qiskit.QuantumRegister(n_qubits)

        anc, circuit = self._initialize_circuit(memory, remain)

        index_differ = 0

        for k in target:
            if index_nonzero[k] != self.index_zero[k]:
                index_differ = k
                ctrl_state = index_nonzero[k]
                break

        target_cx = []
        for k in target:
            if index_differ != k and index_nonzero[k] != self.index_zero[k]:
                circuit.cx(index_differ, k, ctrl_state=ctrl_state)
                target_cx.append(k)

        for k in remain:
            if index_nonzero[k] != self.index_zero[k]:
                circuit.cx(index_differ, k, ctrl_state=ctrl_state)
                target_cx.append(k)

        for k in remain:
            if self.index_zero[k] == '0':
                circuit.x(k)

        if aux:
            # apply mcx using mode v-chain
            mcxvchain(circuit, memory, anc, remain, index_differ)
        else:
            circuit.mcx(remain, index_differ)

        for k in remain:
            if self.index_zero[k] == '0':
                circuit.x(k)

        self.next_state = self._next_state(ctrl_state, index_differ, remain, target_cx)

        return circuit

    def _initialize_circuit(self, memory, remain):
        if self.aux:
            n_anci = len(remain)
            anc = qiskit.QuantumRegister(n_anci - 1, name='anc')
            circuit = qiskit.QuantumCircuit(memory, anc)

        else:
            circuit = qiskit.QuantumCircuit(memory)
            anc = None

        return anc, circuit

    def _get_index_zero(self, non_zero, state):
        index_zero = None
        for k in range(2 ** non_zero):
            txt = '0' + str(self.n_qubits) + 'b'
            index = format(k, txt)

            if index not in state:
                index_zero = index
                break

        return index_zero

    def _get_index_nz(self, target_size):
        index_nonzero = None
        for index, _ in self.next_state.items():
            if index[:target_size] != target_size * '0':
                index_nonzero = index
                break
        return index_nonzero

    @staticmethod
    def initialize(state, aux=False):
        """ Create circuit to initialize a sparse quantum state arXiv:2006.00016

        For instance, to initialize the state a|001>+b|100>
            $ state = {'001': a, '100': b}
            $ circuit = sparse_initialize(state)

        Parameters
        ----------
        state: dict of {str:int}
            A unit vector representing a quantum state.
            Keys are binary strings and values are amplitudes.

        aux: bool
            circuit with auxiliary qubits if aux == True

        Returns
        -------
        sp_circuit: QuantumCircuit
            QuantumCircuit to initialize the state
        """

        return PivotStatePreparation(state, aux).sp_circuit




def mcxvchain(circuit, memory, anc, lst_ctrl, tgt):
    """ multi-controlled x gate with working qubits """
    circuit.rccx(memory[lst_ctrl[0]], memory[lst_ctrl[1]], anc[0])
    for j in range(2, len(lst_ctrl)):
        circuit.rccx(memory[lst_ctrl[j]], anc[j - 2], anc[j - 1])

    circuit.cx(anc[len(lst_ctrl) - 2], tgt)

    for j in reversed(range(2, len(lst_ctrl))):
        circuit.rccx(memory[lst_ctrl[j]], anc[j - 2], anc[j - 1])
    circuit.rccx(memory[lst_ctrl[0]], memory[lst_ctrl[1]], anc[0])



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Bounded Approximation Algorithm.
https://arxiv.org/abs/2111.03132
"""

from dataclasses import dataclass
from itertools import combinations, chain
from typing import List, Optional, Tuple
from math import log2, sqrt
import numpy as np
from tensorly.tenalg.core_tenalg import kronecker

from qclib.entanglement import geometric_entanglement
from qclib.state_preparation.schmidt import cnot_count as schmidt_cnots, \
                                            schmidt_decomposition, \
                                            schmidt_composition, \
                                            low_rank_approximation

# pylint: disable=missing-class-docstring

def adaptive_approximation(state_vector, max_fidelity_loss, strategy='greedy',
                                        max_combination_size=0, use_low_rank=False):
    """
    It reduces the entanglement of the given state, producing an approximation
    to reduce the complexity of the quantum circuit needed to prepare it.
    `https://arxiv.org/abs/2111.03132`_.
    Args:
        state_vector (list):
            A state vector to be approximated by a less complex state.
        max_fidelity_loss (float):
            Maximum fidelity loss allowed to the approximated state.
        strategy (string):
            Method to search for the best approximation ('brute_force' or 'greedy').
            For states larger than 2**8, the greedy strategy should preferably be used.
            Default is strategy='greedy'.
        max_combination_size (int):
            Maximum size of the combination ``C(n_qubits, max_combination_size)``
            between the qubits of an entangled subsystem of length ``n_qubits`` to
            produce the possible bipartitions
            (1 <= ``max_combination_size`` <= ``n_qubits``//2).
            For example, if ``max_combination_size``==1, there will be ``n_qubits``
            bipartitions between 1 and ``n_qubits``-1 qubits.
            The default value is 0 (the size will be maximum for each level).
        use_low_rank (bool):
            If set to True, ``rank``>1 approximations are also considered. This is fine
            tuning for high-entanglement states and is slower.
            The default value is False.
    Returns:
        Node: a node with the data required to build the quantum circuit.
    """

    n_qubits = _to_qubits(len(state_vector))

    # Completely separates the state to estimate the maximum possible fidelity loss.
    # If max_fidelity_loss input is lower than the estimated loss, it runs the full
    # routine with potentially exponential cost.
    entanglement, product_state = geometric_entanglement(state_vector, return_product_state=True)
    if max_fidelity_loss >= entanglement:
        qubits = [(n,) for n in range(n_qubits)]
        ranks = [1] * n_qubits
        partitions = [None] * n_qubits
        cnots = schmidt_cnots(state_vector, method='estimate')
        return Node(
            cnots, cnots, entanglement, entanglement, product_state, qubits, ranks, partitions, []
        )

    vectors = [state_vector]
    qubits = [tuple(range(n_qubits))]
    ranks = [0]
    partitions = [None]

    root_node = Node(0, 0, 0.0, 0.0, vectors, qubits, ranks, partitions, [])
    _build_approximation_tree(root_node, max_fidelity_loss, strategy,
                                    max_combination_size, use_low_rank)

    leaves = []
    _search_leaves(root_node, leaves)

    best_node = _search_best(leaves)

    return best_node

@dataclass
class Entanglement:
    """
    Entanglement reduction information.

    This class contains the information about the entanglement reduction
    of a bipartition. It can be used to assemble an approximate state
    (rank>1) or two completely separate states (rank=1).
    """
    rank: int
    svd_u: np.ndarray
    svd_v: np.ndarray
    svd_s: np.ndarray

    register: Tuple[int]
    partition: Tuple[int]
    local_partition: Tuple[int]

    fidelity_loss: float

@dataclass
class Node:
    """
    Tree node used in _approximation_tree function.
    """
    node_saved_cnots: int
    total_saved_cnots: int

    node_fidelity_loss: float
    total_fidelity_loss: float

    vectors: List[List[complex]]
    qubits: List[Tuple[int]]
    ranks: List[int]
    partitions: List[Optional[Tuple[int]]]

    nodes: List['Node']

    @property
    def is_leaf(self) -> bool:
        """
        True if the all vectors have reached an approximation assessment. There
        is no more decomposition/approximation possible. Therefore, the node is
        a leaf.
        """
        return all(np.asarray(self.ranks) >= 1)

    def num_qubits(self) -> int:
        """ Complete state number of qubits. """
        return len([e for qb_list in self.qubits for e in qb_list])

    def state_vector(self) -> np.ndarray:
        """ Complete state vector. """
        # The vectors are not necessarily in the correct order, but these are
        # given by the qubits field. We need to arrange them so that we have the
        # vectors in the correct ordering!
        # As there are full vectors spanning non-consecutive qubits, we flatten the
        # qubits and deduce from that, how we need to move the axis
        flatten_qubits = [e for q in self.qubits for e in q]

        # The new order is given by ordering the flattened qubits
        new_order = [v[0] for v in sorted(enumerate(flatten_qubits), key=lambda v: v[1])]

        # Reshaping the vector must take the qubit structure into account.
        no_qubits = len(flatten_qubits)
        qubit_shape = [2] * no_qubits
        state = kronecker(self.vectors).reshape(qubit_shape)
        # Moveaxis to the rescue: we now can move the qubits axis and reshape to a vector
        state = np.moveaxis(state, new_order, range(len(new_order))).reshape(-1, )
        return state

    def __str__(self):
        str_vectors = '\n'.join([str(np.around(i,2)) for i in self.vectors])
        str_qubits = ' '.join([str(i) for i in self.qubits])
        str_ranks = ' '.join([str(i) for i in self.ranks])
        return f'saved cnots node={self.node_saved_cnots} ' + \
               f'total={self.total_saved_cnots}\n' + \
               f'fidelity loss node={round(self.node_fidelity_loss,6)} ' + \
               f'total={round(self.total_fidelity_loss,6)}\n' + \
               f'states\n{str_vectors}\n' + \
               f'qubits\n{str_qubits}\n' + \
               f'ranks\n{str_ranks}'

def _build_approximation_tree(node, max_fidelity_loss, strategy='brute_force', max_k=0,
                                                                    use_low_rank=False):
    # Ignore states that are already completely or partially disentangled.
    node_data = [(q, v) for q, v, k in zip(node.qubits, node.vectors, node.ranks) if k == 0]

    for entangled_qubits, entangled_vector in node_data:

        if not 1 <= max_k <= len(entangled_qubits)//2:
            max_k = len(entangled_qubits)//2

        if strategy == 'greedy':
            combs = _greedy_combinations(entangled_vector, entangled_qubits, max_k)
        else:
            combs = _all_combinations(entangled_qubits, max_k)

        # Disentangles or reduces the entanglement of each bipartion of
        # entangled_qubits.
        for partition in combs:
            # Computes the two state vectors after disentangling "partition".
            # If the bipartition cannot be fully disentangled, an approximate
            # state is returned.
            entanglement_info = _reduce_entanglement(
                entangled_vector, entangled_qubits, partition, use_low_rank
            )

            node_fidelity_loss = np.array([e_info.fidelity_loss for e_info in entanglement_info])
            total_fidelity_loss = 1.0 - (1.0 - node_fidelity_loss) * \
                                        (1.0 - node.total_fidelity_loss)

            for e_info, loss in zip(entanglement_info, total_fidelity_loss):
                # Recursion should not continue for this branch if
                # "total_fidelity_loss" has reached "max_fidelity_loss".
                # The leaf corresponds to the node of the best approximation of
                # "max_fidelity_loss" on the branch.
                if loss <= max_fidelity_loss:
                    new_node = _create_node(node, e_info)
                    if new_node.total_saved_cnots > 0:
                        node.nodes.append(new_node)

    if len(node.nodes) > 0:  # If it is not the end of the recursion,
        node.vectors.clear() # clear vectors and qubits to save memory.
        node.qubits.clear()  # This information is no longer needed from this point
                             # on (but may be needed in the future).
    if strategy == 'greedy' and len(node.nodes) > 0:
        # Locally optimal choice at each stage.
        node.nodes = [_search_best(node.nodes)]

    for new_node in node.nodes:
        # call _build_approximation_tree recurrently for each new node.
        if not new_node.is_leaf: # Saves one call for each leaf node.
            _build_approximation_tree(
                new_node, max_fidelity_loss, strategy, max_k, use_low_rank
            )

def _all_combinations(entangled_qubits, max_k):
    combs = tuple(combinations(entangled_qubits, max_k))
    if len(entangled_qubits)%2 == 0 and len(entangled_qubits)//2 == max_k:
        # Ignore redundant complements. Only when max_k is exactly
        # half the length of entangled_qubits. Reduces the number of branches.
        # (0,1,2,3) -> (0,1), (0,2), (0,3); ignore (2,3), (1,3), (1,2) .
        combs = combs[:len(combs)//2]

    return chain(*(combinations(entangled_qubits, k)
                                            for k in range(1, max_k)), combs)

def _greedy_combinations(entangled_vector, entangled_qubits, max_k):
    """
    Combinations with a qubit-by-qubit analysis.
    Returns only one representative of the partitions of size k (1<=k<=max_k).
    The increment in the partition size is done by choosing the qubit that has
    the lowest fidelity-loss when removed from the remaining entangled subsystem.
    """
    node = Node( 0, 0, 0.0, 0.0, [entangled_vector], [entangled_qubits], [0], [None], [] )
    for _ in range(max_k):
        current_vector = node.vectors[-1] # Last item is the current entangled state.
        current_qubits = node.qubits[-1]

        nodes = []
        # Disentangles one qubit at a time.
        for qubit_to_disentangle in current_qubits:
            entanglement_info = \
                _reduce_entanglement(current_vector, current_qubits, (qubit_to_disentangle,))

            new_node = _create_node(node, entanglement_info[0])

            nodes.append(new_node)
        # Search for the node with lowest fidelity-loss.
        node = _search_best(nodes)

    # Build the partitions by incrementing the number of selected qubits.
    # Returns only one partition for each length k.
    # All disentangled qubits are in the slice "node.qubits[0:max_k]", in the order in which
    # they were selected. Each partition needs to be sorted to ensure that the correct
    # construction of the circuit.
    return ( tuple(sorted( chain(*node.qubits[:k]) )) for k in range(1, max_k+1) )

def _reduce_entanglement(state_vector, register, partition, use_low_rank=False):
    local_partition = []
    # Maintains the relative position between the qubits of the two subsystems.
    for qubit_to_disentangle in partition:
        local_partition.append(sum(i < qubit_to_disentangle for i in register))

    local_partition = tuple(local_partition)

    svd_u, svd_s, svd_v = schmidt_decomposition(state_vector, local_partition)

    entanglement_info = []

    max_ebits = 0
    if use_low_rank:
        # Limit the maximum low_rank to "2**(total_ebits-1)" to not repeat the original state.
        max_ebits = _to_qubits(svd_s.shape[0]) - 1

    for ebits in range(0, max_ebits+1):
        low_rank = 2**ebits

        rank, low_rank_u, low_rank_v, low_rank_s = \
            low_rank_approximation(low_rank, svd_u, svd_v, svd_s)

        if rank < low_rank:
            break # No need to go any further, as the maximum effective rank has been reached.

        fidelity_loss = 1.0 - sum(low_rank_s**2)

        entanglement_info.append(Entanglement(rank, low_rank_u, low_rank_v, low_rank_s,
                                              register,
                                              partition,
                                              local_partition,
                                              fidelity_loss))
    return entanglement_info

def _create_node(parent_node, e_info):

    vectors = parent_node.vectors.copy()
    qubits = parent_node.qubits.copy()
    ranks = parent_node.ranks.copy()
    partitions = parent_node.partitions.copy()

    index = parent_node.qubits.index(e_info.register)
    original_vector = vectors.pop(index)
    original_qubits = qubits.pop(index)
    original_rank = ranks.pop(index)
    original_partition = partitions.pop(index)

    if e_info.rank == 1:
        # The partition qubits have been completely disentangled from the
        # rest of the register. Therefore, the original entangled state is
        # removed from the list and two new separate states are included.
        partition1 = tuple( sorted(set(original_qubits).difference(set(e_info.partition))) )
        partition2 = e_info.partition

        vectors.append(e_info.svd_v.T[:, 0])
        qubits.append(partition2)
        ranks.append(1 if len(partition2) == 1 else 0) # Single qubit states can
        partitions.append(None)                        # no longer be disentangled.

        vectors.append(e_info.svd_u[:, 0])
        qubits.append(partition1)
        ranks.append(1 if len(partition1) == 1 else 0)
        partitions.append(None)

        node_saved_cnots = _count_saved_cnots(
                                original_vector, vectors[-1], vectors[-2],
                                original_partition, None,
                                original_rank
                            )
    else:
        # The entanglement between partition qubits and the rest of the
        # register has been reduced, but not eliminated. Therefore, the
        # original state is replaced by an approximate state.
        normed_svd_s = e_info.svd_s/sqrt( 1.0 - e_info.fidelity_loss )
        approximate_state = schmidt_composition(e_info.svd_u, e_info.svd_v,
                                                normed_svd_s, e_info.local_partition)
        vectors.append(approximate_state)
        qubits.append(original_qubits)
        ranks.append(e_info.rank)
        partitions.append(e_info.local_partition)

        node_saved_cnots = _count_saved_cnots(
                                original_vector, vectors[-1], None,
                                original_partition, e_info.local_partition,
                                original_rank, e_info.rank
                            )

    total_saved_cnots = parent_node.total_saved_cnots + node_saved_cnots
    total_fidelity_loss = 1.0 - (1.0 - e_info.fidelity_loss) * \
                                (1.0 - parent_node.total_fidelity_loss)

    return Node(node_saved_cnots, total_saved_cnots, e_info.fidelity_loss,
                    total_fidelity_loss, vectors, qubits, ranks, partitions, [])

def _search_leaves(node, leaves):
    # It returns the leaves of the tree. These nodes are the ones with
    # total_fidelity_loss closest to max_fidelity_loss for each branch.
    if len(node.nodes) == 0:
        leaves.append(node)
    else:
        for child in node.nodes:
            _search_leaves(child, leaves)

def _search_best(nodes):
    # Nodes with the greatest reduction in the number of CNOTs.
    # There may be several with the same number.
    max_total_saved_cnots = max(nodes, key=lambda n: n.total_saved_cnots).total_saved_cnots
    max_saved_cnots_nodes = [node for node in nodes
                                if node.total_saved_cnots == max_total_saved_cnots]
    # Nodes with the minimum depth (wich depends on the size of the node's largest subsystem).
    # Shallower circuits with the same number of CNOTs means more parallelism.
    min_depth = _max_subsystem_size(min(max_saved_cnots_nodes, key=_max_subsystem_size))
    min_depth_nodes = [node for node in max_saved_cnots_nodes
                                if _max_subsystem_size(node) == min_depth]
    # Node with the lowest fidelity loss among the nodes with
    # the highest reduction in the number of CNOTs.
    return min(min_depth_nodes, key=lambda n: n.total_fidelity_loss)

def _max_subsystem_size(node):
    return len(max(node.qubits, key=len))

def _to_qubits(n_state_vector):
    return int(log2(n_state_vector))

def _count_saved_cnots(original_vector, subsystem1_vector, subsystem2_vector,
                            original_partition=None,subsystem_local_partition=None,
                            original_rank=0, subsystem_rank=0):
    method = 'estimate'

    cnots_originally = schmidt_cnots(
                        original_vector, method=method, partition=original_partition,
                        low_rank=original_rank
                    )
    cnots_phase_3 = schmidt_cnots(
                        subsystem1_vector, method=method, partition=subsystem_local_partition,
                        low_rank=subsystem_rank
                    )

    cnots_phase_4 = 0
    if subsystem2_vector is not None:
        cnots_phase_4 = schmidt_cnots(subsystem2_vector, method=method)

    return cnots_originally - cnots_phase_3 - cnots_phase_4



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the mmes.py module.
"""

from unittest import TestCase
from itertools import combinations
from math import factorial
import numpy as np
import qutip
from qclib.state_preparation.util.mmes import create_state

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

class TestMmes(TestCase):
    def _nCr(self, n,r):
        # pylint: disable=invalid-name
        return factorial(n) // factorial(r) // factorial(n-r)

    def test_state_creation(self):
        for n_qubits in range(5, 6):
            state = create_state(n_qubits)
            #                                10                     20                        30
            #        0,1,2,3,4, 5, 6,7,8, 9, 0, 1,2,3,4,5,6,7, 8, 9,0, 1,2, 3, 4,5, 6,7, 8, 9,0,1
            #state = [0,0,0,0,0, 1,-1,0,1, 0, 0,-1,0,0,0,0,0,1, 1, 0,0, 0,0, 0, 1,0, 0,0, 0, 0,0,1 ]
            state = [1,1,1,1,1,-1,-1,1,1,-1,-1, 1,1,1,1,1,1,1,-1,-1,1,-1,1,-1,-1,1,-1,1,-1,-1,1,1]
            state = state/np.linalg.norm(state)

            print(state)
            ket = qutip.Qobj(state, dims=[[2] * (n_qubits), [1] * (n_qubits)]).unit()

            pi_a_sum = 0                    # sum of purity of the subsystems
            n_a = n_qubits//2
            combs = combinations(range(n_qubits), n_a)
            for subsystem_a in combs:
                rho_sq = ket.ptrace(subsystem_a)**2
                pi_a = rho_sq.tr()          # purity of subsystem_a
                pi_a_sum += pi_a
                print(subsystem_a)
                print(pi_a)
                self.assertTrue(1/2**n_a<=pi_a<=1)

            # potential of multipartite entanglement measures the average
            # bipartite entanglement over all possible balanced bipartition
            # 1/2^n_a <= pi_me <= 1.
            print('nCr', self._nCr(n_qubits, n_a))
            pi_me = ( 1/self._nCr(n_qubits, n_a) ) * pi_a_sum
            print('pi_a_sum', pi_a_sum)
            print('pi_me', pi_me)
            self.assertTrue(1/2**n_a<=pi_me<=1)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test bidirectional state preparation """

from unittest import TestCase
import numpy as np
from qiskit import ClassicalRegister
from qiskit.providers.aer.backends import AerSimulator
from qclib.state_preparation.bdsp import initialize
from .util import measurement

backend = AerSimulator()
SHOTS = 8192


class TestBdsp(TestCase):
    """ Testing bdsp """

    @staticmethod
    def bdsp_experiment(state, split=None):
        """ Run bdsp experiment """
        circuit = initialize(state, split)

        n_qubits = int(np.log2(len(state)))
        classical_register = ClassicalRegister(n_qubits)
        circuit.add_register(classical_register)

        return measurement(circuit, n_qubits, classical_register, backend, SHOTS)

    def test_bottom_up(self):
        """ Testing bdsp """

        vector = np.random.rand(16) + np.random.rand(16) * 1j
        vector = vector / np.linalg.norm(vector)

        state = TestBdsp.bdsp_experiment(vector, 1)

        self.assertTrue(np.allclose(np.power(np.abs(vector), 2), state, rtol=1e-01, atol=0.005))

    def test_top_down(self):
        """ Testing bdsp """
        vector = np.random.rand(16) + np.random.rand(16) * 1j
        vector = vector / np.linalg.norm(vector)

        state = TestBdsp.bdsp_experiment(vector, int(np.log2(len(vector))))

        self.assertTrue(np.allclose(np.power(np.abs(vector), 2), state, rtol=1e-01, atol=0.005))

    def test_sublinear(self):
        """ Testing bdsp """
        vector = np.random.rand(16) + np.random.rand(16) * 1j
        vector = vector / np.linalg.norm(vector)

        state = TestBdsp.bdsp_experiment(vector)

        self.assertTrue(np.allclose(np.power(np.abs(vector), 2), state, rtol=1e-01, atol=0.005))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the schmidt.py module.
"""

from unittest import TestCase
import numpy as np
from itertools import combinations
from qiskit import ClassicalRegister, execute, transpile
from qiskit.providers.aer.backends import AerSimulator
from qclib.state_preparation.schmidt import initialize
from qclib.util import get_state

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

class TestSchmidt(TestCase):
    @staticmethod
    def mae(state, ideal):
        """
        Mean Absolute Error
        """
        return np.sum(np.abs(state-ideal))/len(ideal)

    @staticmethod
    def fidelity(state1, state2):
        bra = np.conj(state1)
        ket = state2

        return np.abs(bra.dot(ket))**2

    @staticmethod
    def get_counts(circuit):
        n_qubits = circuit.num_qubits
        classical_reg = ClassicalRegister(n_qubits)
        circuit.add_register(classical_reg)
        circuit.measure(list(range(n_qubits)), classical_reg)

        backend = AerSimulator()
        counts = execute(circuit, backend, shots=8192).result().get_counts()

        counts_with_zeros = {}
        for i in range(2**n_qubits):
            pattern = '{:0{}b}'.format(i, n_qubits)
            if pattern in counts:
                counts_with_zeros[pattern] = counts[pattern]
            else:
                counts_with_zeros[pattern] = 0.0

        sum_values = sum(counts.values())
        return [ value/sum_values for (key, value) in counts_with_zeros.items() ]

    def _test_initialize_mae(self, rank=0, max_mae=10**-15):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        qubits = list(range(5))
        partitions = combinations(qubits, 3)
        for partition in partitions:
            circuit = initialize(state_vector, low_rank=rank, partition=partition)

            state = get_state(circuit)

            self.assertTrue(TestSchmidt.mae(state, state_vector) < max_mae)

    def test_initialize_full_rank(self):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state))

    def test_initialize_ame(self):
        """ Test initialization of a absolutely maximally entangled state"""
        state_vector = [1, 1, 1, 1,1,-1,-1, 1, 1,-1,-1, 1, 1, 1,1,1,
                        1, 1,-1,-1,1,-1, 1,-1,-1, 1,-1, 1,-1,-1,1,1]
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state))

    def test_measurement_full_rank(self):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector)

        state = TestSchmidt.get_counts(circuit)

        self.assertTrue(np.allclose( np.power(np.abs(state_vector),2), state,
                        rtol=1e-01, atol=0.005))

    def test_initialize_rank_5(self):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = initialize(state_vector, low_rank=5)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state)) # same as unitary.

    def test_initialize_rank_4(self):
        self._test_initialize_mae(4, max_mae=10**-14)

    def test_initialize_rank_3(self):
        self._test_initialize_mae(3, max_mae=0.04)

    def test_initialize_rank_2(self):
        self._test_initialize_mae(2, max_mae=0.06)

    def test_initialize_rank_1(self):
        self._test_initialize_mae(1, max_mae=0.09)

    def test_cnot_count_rank_1(self):
        # Builds a rank 1 state.
        state_vector = [1]
        for _ in range(5):
            vec = np.random.rand(2) + np.random.rand(2) * 1j
            vec = vec / np.linalg.norm(vec)
            state_vector = np.kron(state_vector, vec)

        circuit = initialize(state_vector)
        transpiled_circuit = transpile(circuit, basis_gates=['u', 'cx'], optimization_level=3)

        # The cnots count must be zero (i.e. not present in the dictionary).
        self.assertTrue('cx' not in transpiled_circuit.count_ops())



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Bounded Approximation version of Plesch's algorithm.
https://arxiv.org/abs/2111.03132
"""

import numpy as np
from qiskit import QuantumCircuit
from qclib.state_preparation import schmidt
from qclib.state_preparation.util.baa import adaptive_approximation

def initialize(state_vector, max_fidelity_loss=0.0,
                        isometry_scheme='ccd', unitary_scheme='qsd',
                        strategy='greedy', max_combination_size=0,
                        use_low_rank=False, return_node=False):
    """
    State preparation using the bounded approximation algorithm via Schmidt
    decomposition arXiv:1003.5760

    For instance, to initialize the state a|00> + b|10> (|a|^2+|b|^2=1)
        $ state = [a, 0, b, 0]
        $ circuit = initialize(state)

    Parameters
    ----------
    state_vector: list of float or array-like
        A unit vector representing a quantum state.
        Values are amplitudes.

    max_fidelity_loss: float
        ``state`` allowed (fidelity) error for approximation (0 <= ``max_fidelity_loss`` <= 1).
        If ``max_fidelity_loss`` is not in the valid range, it will be ignored.

    isometry_scheme: string
        Scheme used to decompose isometries.
        Possible values are ``'knill'`` and ``'ccd'`` (column-by-column decomposition).
        Default is ``isometry_scheme='ccd'``.

    unitary_scheme: string
        Scheme used to decompose unitaries.
        Possible values are ``'csd'`` (cosine-sine decomposition) and ``'qsd'`` (quantum
        Shannon decomposition).
        Default is ``unitary_scheme='qsd'``.

    strategy: string
        Method to search for the best approximation (``'brute_force'`` or ``'greedy'``).
        For states larger than 2**8, the greedy strategy should preferably be used.
        Default is ``strategy='greedy'``.

    max_combination_size: int
        Maximum size of the combination ``C(n_qubits, max_combination_size)``
        between the qubits of an entangled subsystem of length ``n_qubits`` to
        produce the possible bipartitions
        (1 <= ``max_combination_size`` <= ``n_qubits``//2).
        For example, if ``max_combination_size``==1, there will be ``n_qubits``
        bipartitions between 1 and ``n_qubits``-1 qubits.
        The default value is 0 (the size will be maximum for each level).

    use_low_rank: bool
            If set to True, ``rank``>1 approximations are also considered. This is fine
            tuning for high-entanglement states and is slower.
            The default value is False.

    return_node: bool
            If set to true, returns also the best node for the
            decomposition/approximation

    Returns
    -------
    circuit: QuantumCircuit or Tuple[QuantumCircuit, Node]
        QuantumCircuit to initialize the state or if return_node==True, returns
        a tuple with the QuantumCircuit and the best node for
        decomposition/approximation.
    """

    if max_fidelity_loss < 0 or max_fidelity_loss > 1:
        max_fidelity_loss = 0.0

    node = adaptive_approximation(
        state_vector, max_fidelity_loss, strategy, max_combination_size, use_low_rank
    )

    n_qubits = int(np.log2(len(state_vector)))
    circuit = QuantumCircuit(n_qubits)

    for vector, qubits, rank, partition in zip(node.vectors, node.qubits,
                                                node.ranks, node.partitions):
        qc_vec = schmidt.initialize(vector, rank, isometry_scheme, unitary_scheme, partition)
        circuit.compose(qc_vec, qubits[::-1], inplace=True) # qiskit little-endian.

    if return_node:
        return circuit.reverse_bits(), node

    return circuit.reverse_bits()



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
https://arxiv.org/abs/2108.10182
"""


from dataclasses import dataclass
import numpy as np

from qclib.state_preparation.util.tree_utils import is_leaf

@dataclass
class NodeAngleTree:
    """
    Binary tree node used in function create_angles_tree
    """
    index: int
    level: int
    angle_y: float
    angle_z: float
    left: 'NodeAngleTree'
    right: 'NodeAngleTree'

    def __str__(self):
        return str(self.level) + '_' + \
               str(self.index) + '\n' + \
               '{0:.2g}'.format(self.angle_y) + '_' + \
               '{0:.2g}'.format(self.angle_z)

def create_angles_tree(state_tree):
    """
    :param state_tree: state_tree is an output of state_decomposition function
    :param tree: used in the recursive calls
    :return: tree with angles that will be used to perform the state preparation
    """
    amp = 0.0
    if state_tree.amplitude != 0.0:
        amp = state_tree.right.amplitude / state_tree.amplitude

    # Avoid out-of-domain value due to numerical error.
    if np.abs(amp) < -1.0:
        angle_y = -np.pi
    elif np.abs(amp) > 1.0:
        angle_y = np.pi
    else:
        angle_y = 2 * np.arcsin( np.abs(amp) )

    angle_z = 2 * np.angle(amp)

    node = NodeAngleTree(state_tree.index, state_tree.level, angle_y, angle_z, None, None)

    if not is_leaf(state_tree.left):
        node.right = create_angles_tree(state_tree.right)
        node.left = create_angles_tree(state_tree.left)

    return node



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
n-qubit controlled gate
"""
import qiskit
import numpy as np


def mc_gate(gate:np.ndarray, qcirc: qiskit.QuantumCircuit, controls: list, targ: int):
    """

    Parameters
    ----------
    qcirc: qiskit.QuantumCircuit
    controls: list of control qubits
    targ: target qubit

    Returns
    -------

    """

    n_qubits = len(controls) + 1
    gate_circuit = qiskit.QuantumCircuit(n_qubits, name="T" + str(targ))
    coef = _coefficients(n_qubits)

    c1c2(gate, coef, n_qubits, gate_circuit)
    _c3(gate, coef, n_qubits, gate_circuit)

    c1c2(gate, coef[:-1,1:-1], n_qubits-1, gate_circuit, False)
    _c3(gate, coef[:-1, 1:-1], n_qubits - 1, gate_circuit, False)

    qcirc.compose(gate_circuit, controls + [targ], inplace=True)


def _c3(gate, coef, n_qubits, qcirc, first=True):
    """ phase 2 """
    for k in range(1, n_qubits-1):
        line = k
        column = n_qubits - 1

        while column < 2*n_qubits-3 and coef[line, column] != 0:
            control = (column % (n_qubits - 1) + 1)
            targ = line + 1
            # -----
            if first and targ == n_qubits-1:
                csqgate = _xgate(gate, coef, column, line)
                qcirc.compose(csqgate, qubits=[control, targ], inplace=True)
            else:
                qcirc.crx(np.pi / coef[line, column], control, targ)

            line = line - 1
            column = column + 1

    for k in range(n_qubits, 2*n_qubits-3):
        line = n_qubits-2
        column = k
        control = column - line
        while column< 2*n_qubits-3 and coef[line, column] != 0:
            targ = line + 1
            # -----
            if first and targ == n_qubits - 1:
                csqgate = _xgate(gate, coef, column, line)
                qcirc.compose(csqgate, qubits=[control, targ], inplace=True)
            else:
                qcirc.crx(np.pi / coef[line, column], control, targ)

            line = line - 1
            column = column + 1
            control = control + 1


def c1c2(gate, coef, n_qubits, qcirc, first=True):
    """ phase 1 """
    if not first:
        for k in range(n_qubits-1):
            coef[k, n_qubits-2] = - coef[k, n_qubits-2]

    for k in range(n_qubits - 1):
        line = n_qubits - 2
        column = k
        while line >= 0 and column >= 0 and coef[line, column] != 0:
            control = n_qubits - 2 - column
            targ = line + 1
            # -----
            if first and targ == n_qubits - 1:
                csqgate = _xgate(gate, coef, column, line)
                qcirc.compose(csqgate, qubits=[control, targ], inplace=True)
            else:
                qcirc.crx(np.pi / coef[line, column], control, targ)

            line = line - 1
            column = column - 1



    for k in range(n_qubits - 3, -1, -1):
        column = n_qubits - 2
        line = k


        while line >= 0 and column >= 0 and coef[line, column] != 0:
            # -----
            control = n_qubits - 2 - column
            targ = line + 1
            if first and targ == n_qubits - 1:
                csqgate = _xgate(gate, coef, column, line)
                qcirc.compose(csqgate, qubits=[control, targ], inplace=True)
            else:
                qcirc.crx(np.pi / coef[line, column], control, targ)

            line = line - 1
            column = column - 1


def _xgate(agate, coef, column, line):
    signal = np.sign(coef[line, column])
    param = np.abs(coef[line, column])

    values, vectors = np.linalg.eig(agate)
    gate = np.power(values[0]+0j, 1/param) * vectors[:,[0]] @ vectors[:,[0]].conj().T + \
           np.power(values[1]+0j, 1/param) * vectors[:,[1]] @ vectors[:,[1]].conj().T

    if signal < 0:
        gate = np.linalg.inv(gate)

    sqgate = qiskit.QuantumCircuit(1, name='X^1/' + str(coef[line, column]))
    sqgate.unitary(gate, 0) #pylint: disable=maybe-no-member
    csqgate = sqgate.control(1)

    return csqgate


def _coefficients(n_qubits):
    coef = np.zeros((n_qubits - 1, 2 * n_qubits - 3))

    for i in range(0, n_qubits - 2):
        one_coef = 2 ** (i+1)

        for j in range(n_qubits - 2, -1, -1):
            coef[j, i] = one_coef
            coef[j, (-i + 2 * n_qubits - 4)] = -one_coef
            one_coef = one_coef // 2

            if one_coef < 2:
                break

    for k in range(n_qubits - 1):
        coef[k][n_qubits - 2] = 2 ** k

    return coef

#
# if __name__ == '__main__':
#     from qclib.util import get_counts, get_state
#
#
#     qcirc = qiskit.QuantumCircuit(6)
#     # qcirc.x(0)
#     qcirc.x(1)
#     qcirc.x(2)
#     # qcirc.x(3)
#     # qcirc.x(4)
#
#
#     toffoli(qcirc, [0, 1, 2, 3, 4], 5)
#     qcirc.measure_all()
#     print(get_counts(qcirc))
#     print(1)
